from __future__ import annotations

import logging
import re
from datetime import datetime

import matplotlib.colors as mcolors
import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from iminuit import Minuit, cost
from matplotlib.colors import LogNorm
from scipy.stats import linregress

import pygama.math.histogram as pgh
import pygama.pargen.AoE_cal as AoE
from pygama.math.distributions import gaussian
from pygama.pargen.survival_fractions import compton_sf_sweep, get_sf_sweep

log = logging.getLogger(__name__)


def get_fit_range(lq: np.array) -> tuple(float, float):
    """
    Function for determining the fit range for a given distribution of lq values
    """

    # Get an initial guess of mu and sigma, use these values to determine our final fit range
    left_guess = np.nanpercentile(lq, 1)
    right_guess = np.nanpercentile(lq, 95)
    test_range = (left_guess, right_guess)

    hist, bins, _ = pgh.get_hist(lq, bins=100, range=test_range)
    bin_centers = (bins[:-1] + bins[1:]) / 2
    mu = bin_centers[np.argmax(hist)]
    _, sigma, _ = pgh.get_gaussian_guess(hist, bins)

    left_edge = mu - 2.5 * sigma
    right_edge = mu + 2.5 * sigma
    fit_range = (left_edge, right_edge)

    return fit_range


def get_lq_hist(
    df: pd.DataFrame(),
    lq_param: str,
    cal_energy_param: str,
    peak: float,
    sidebands: bool = True,
):
    """
    Function for getting a distribution of LQ values for a given peak. Returns a histogram of the
    LQ distribution as well as an array of bin edges
    """

    if sidebands:
        # Get a histogram of events in the peak using sideband subtraction
        # Uses a 13 keV window, and the sideband is to the right of the peak
        # Default option

        df_peak = df.query(
            f"{cal_energy_param} < ({peak} + 5) & {cal_energy_param} > ({peak} - 8)"
        )
        df_sideband = df.query(
            f"{cal_energy_param} < ({peak} + 20) & {cal_energy_param} > ({peak} + 7)"
        )

        fit_range = get_fit_range(df_peak[lq_param].to_numpy())

        sideband_hist, bins, _ = pgh.get_hist(
            df_sideband[lq_param].to_numpy(), bins=100, range=fit_range
        )
        dep_hist, _, _ = pgh.get_hist(
            df_peak[lq_param].to_numpy(), bins=100, range=fit_range
        )
        final_hist = dep_hist - sideband_hist
        var = np.sqrt(np.add(sideband_hist, dep_hist))

        return final_hist, bins, var

    else:
        # Return a histogram in a 13 keV range surrounding the specified peak

        df_peak = df.query(
            f"{cal_energy_param} < ({peak} + 5) & {cal_energy_param} > ({peak} - 8)"
        )

        fit_range = get_fit_range(df_peak[lq_param].to_numpy())
        dep_hist, bins, var = pgh.get_hist(
            df_peak[lq_param].to_numpy(), bins=100, range=fit_range
        )

        return dep_hist, bins, var


def binned_lq_fit(
    df: pd.DataFrame,
    lq_param: str,
    cal_energy_param: str,
    peak: float,
    cdf=gaussian,
    sidebands: bool = True,
):
    """Function for fitting a distribution of LQ values within a specified
        energy peak. Fits a gaussian to the distribution

    Parameters
    ----------
    df: pd.DataFrame()
        Dataframe containing the data for fitting. Data must
        contain the desired lq parameter and the calibrated
        energy
    lq_param: string
        Name of the LQ parameter to fit
    cal_energy_param: string
        Name of the calibrated energy parameter of choice
    peak: float
        Energy value, in keV, of the peak who's LQ
        distribution will be fit
    cdf: callable
        Function to be used for the binned fit
    sidebands: bool
        Whether or not to perform a sideband subtraction when
        fitting the LQ distribution

    Returns
    -------
    m1.values: array-like object
        Resulting parameter values from the peak fit
    m1.errors: array-like object
        Resulting parameter errors from the peak fit
    hist: array
        Histogram that was used for the binned fit
    bins: array
        Array of bin edges used for the binned fit
    """

    hist, bins, var = get_lq_hist(df, lq_param, cal_energy_param, peak, sidebands)

    # Temporary fix for negative bin counts
    # TODO: Adjust fitting to handle negative bin counts
    hist[hist < 0] = 0

    bin_centers = (bins[:-1] + bins[1:]) / 2

    mu = bin_centers[np.argmax(hist)]
    _, sigma, _ = pgh.get_gaussian_guess(hist, bins)

    c1 = cost.BinnedNLL(hist, bins, gaussian.get_cdf, verbose=0)
    m1 = Minuit(c1, mu=mu, sigma=sigma)
    m1.simplex().migrad()
    m1.hesse()

    return m1.values, m1.errors, hist, bins


def calculate_time_means(
    df: pd.DataFrame,
    lq_param: str,
    cal_energy_param: str,
    peak: float,
    sidebands: bool = True,
):
    """
    Function for calculating the arithmetic mean and sigma for LQ events in a
    specified peak
    """

    df_peak = df.query(
        f"{cal_energy_param} < ({peak} + 5) & {cal_energy_param} > ({peak} - 8)"
    )
    fit_range = get_fit_range(df_peak[lq_param].to_numpy())

    lq_peak_vals = df_peak.query(
        f"{lq_param} < {fit_range[1]} & {lq_param} > {fit_range[0]}"
    )[lq_param].to_numpy()

    mean = np.mean(lq_peak_vals)
    sigma = np.std(lq_peak_vals)
    mean_err = sigma / np.sqrt(len(lq_peak_vals))
    sig_err = 2 * sigma**4 / (len(lq_peak_vals) - 1)

    pars = {"mu": mean, "sigma": sigma}
    errors = {"mu": mean_err, "sigma": sig_err}

    return pars, errors


def fit_time_means(tstamps, means, reses):
    out_dict = {}
    current_tstamps = []
    current_means = []
    current_reses = []

    rolling_mean = means[np.where(~np.isnan(means))[0][0]]
    for i, tstamp in enumerate(tstamps):
        if np.isnan(means[i]) or np.isnan(reses[i]):
            out_dict[tstamp] = np.nan

        else:
            current_tstamps.append(tstamp)
            current_means.append(means[i])
            current_reses.append(reses[i])
            rolling_mean = np.average(
                current_means, weights=1 / np.array(current_reses)
            )
    for tstamp in current_tstamps:
        out_dict[tstamp] = rolling_mean
    return out_dict


class LQCal:
    """A class for calibrating the LQ parameter and determining the LQ cut value"""

    def __init__(
        self,
        cal_dicts: dict,
        cal_energy_param: str,
        dt_param: str,
        eres_func: callable,
        cdf: callable = gaussian,
        selection_string: str = "is_valid_cal&is_not_pulser",
        debug_mode=False,
    ):
        """
        Parameters
        ----------
        cal_dicts: dict
            A dictionary containing the hit-level operations to apply
            to the data.
        cal_energy_param: string
            The calibrated energy parameter of choice
        dt_param: string
            The drift-time parameter of choice
        eres_function: callable
            The energy resolutions function
        cdf: callable
            The CDF used for the binned fits
        selection_string: string
            A string of flags to apply the data when running the calibration
        debug_mode: boolean
            Determines if the class is initialized in debug mode
        """

        self.cal_dicts = cal_dicts
        self.cal_energy_param = cal_energy_param
        self.dt_param = dt_param
        self.eres_func = eres_func
        self.cdf = cdf
        self.selection_string = selection_string
        self.debug_mode = debug_mode

    def update_cal_dicts(self, update_dict):
        if re.match(r"(\d{8})T(\d{6})Z", list(self.cal_dicts)[0]):
            for tstamp in self.cal_dicts:
                if tstamp in update_dict:
                    self.cal_dicts[tstamp].update(update_dict[tstamp])
                else:
                    self.cal_dicts[tstamp].update(update_dict)
        else:
            self.cal_dicts.update(update_dict)

    def lq_timecorr(self, df, lq_param, output_name="LQ_Timecorr", display=0):
        """
        Calculates the average LQ value for DEP events for each specified
        run_timestamp. Applies a time normalization based on the average LQ value
        in the DEP across all run_timestamps.
        """

        log.info("Starting LQ time correction")
        self.timecorr_df = pd.DataFrame()
        try:
            if "run_timestamp" in df:
                for tstamp, time_df in df.groupby("run_timestamp", sort=True):
                    try:
                        pars, errs = calculate_time_means(
                            time_df.query(f"{self.selection_string}"),
                            lq_param,
                            self.cal_energy_param,
                            peak=1592.5,
                        )
                        self.timecorr_df = pd.concat(
                            [
                                self.timecorr_df,
                                pd.DataFrame(
                                    [
                                        {
                                            "run_timestamp": tstamp,
                                            "mean": pars["mu"],
                                            "mean_err": errs["mu"],
                                            "res": pars["sigma"] / pars["mu"],
                                            "res_err": (pars["sigma"] / pars["mu"])
                                            * np.sqrt(
                                                errs["sigma"] / pars["sigma"]
                                                + errs["mu"] / pars["mu"]
                                            ),
                                        }
                                    ]
                                ),
                            ]
                        )
                    except BaseException as e:
                        if e == KeyboardInterrupt:
                            raise (e)
                        elif self.debug_mode:
                            raise (e)

                        self.timecorr_df = pd.concat(
                            [
                                self.timecorr_df,
                                pd.DataFrame(
                                    [
                                        {
                                            "run_timestamp": tstamp,
                                            "mean": np.nan,
                                            "mean_err": np.nan,
                                            "res": np.nan,
                                            "res_err": np.nan,
                                        }
                                    ]
                                ),
                            ]
                        )

                self.timecorr_df.set_index("run_timestamp", inplace=True)
                time_dict = fit_time_means(
                    np.array(self.timecorr_df.index),
                    np.array(self.timecorr_df["mean"]),
                    np.array(self.timecorr_df["res"]),
                )

                df[output_name] = df[lq_param] / np.array(
                    [time_dict[tstamp] for tstamp in df["run_timestamp"]]
                )
                self.update_cal_dicts(
                    {
                        tstamp: {
                            output_name: {
                                "expression": f"{lq_param}/a",
                                "parameters": {"a": t_dict},
                            }
                        }
                        for tstamp, t_dict in time_dict.items()
                    }
                )
                log.info("LQ time correction finished")
            else:
                try:
                    pars, errs = calculate_time_means(
                        df.query(f"{self.selection_string}"),
                        lq_param,
                        self.cal_energy_param,
                        peak=1592.5,
                    )
                    self.timecorr_df = pd.concat(
                        [
                            self.timecorr_df,
                            pd.DataFrame(
                                [
                                    {
                                        "run_timestamp": np.nan,
                                        "mean": pars["mu"],
                                        "mean_err": errs["mu"],
                                        "res": pars["sigma"] / pars["mu"],
                                        "res_err": (pars["sigma"] / pars["mu"])
                                        * np.sqrt(
                                            errs["sigma"] / pars["sigma"]
                                            + errs["mu"] / pars["mu"]
                                        ),
                                    }
                                ]
                            ),
                        ]
                    )
                except BaseException as e:
                    if e == KeyboardInterrupt:
                        raise (e)
                    elif self.debug_mode:
                        raise (e)
                    self.timecorr_df = pd.concat(
                        [
                            self.timecorr_df,
                            pd.DataFrame(
                                [
                                    {
                                        "run_timestamp": np.nan,
                                        "mean": np.nan,
                                        "mean_err": np.nan,
                                        "res": np.nan,
                                        "res_err": np.nan,
                                    }
                                ]
                            ),
                        ]
                    )
                df[output_name] = df[lq_param] / pars["mu"]
                self.update_cal_dicts(
                    {
                        output_name: {
                            "expression": f"{lq_param}/a",
                            "parameters": {"a": pars["mu"]},
                        }
                    }
                )
                log.info("LQ time correction finished")
        except BaseException as e:
            if e == KeyboardInterrupt:
                raise (e)
            elif self.debug_mode:
                raise (e)
            log.error("LQ time correction failed")
            self.update_cal_dicts(
                {
                    output_name: {
                        "expression": f"{lq_param}/a",
                        "parameters": {"a": np.nan},
                    }
                }
            )

    def drift_time_correction(
        self, df: pd.DataFrame(), lq_param, cal_energy_param: str, display: int = 0
    ):
        """
        Deterimines the drift time correction parameters for LQ by fitting a degree 1 polynomial to
        the LQ vs drift time distribution for DEP events. Corrects for any linear dependence and
        centers the final LQ distribution to a mean of 0.
        """

        log.info("Starting LQ drift time correction")
        try:
            pars = binned_lq_fit(df, lq_param, self.cal_energy_param, peak=1592.5)[0]
            mean = pars[0]
            sigma = pars[1]

            dep_events = df.query(
                f"{self.cal_energy_param} > 1589.5 & {self.cal_energy_param} < 1595.5 & {self.cal_energy_param}=={self.cal_energy_param}&{lq_param}=={lq_param}"
            )

            dt_range = [
                np.nanpercentile(dep_events[self.dt_param], 10),
                np.nanpercentile(dep_events[self.dt_param], 95),
            ]

            lq_range = [mean - 2 * sigma, mean + 2 * sigma]

            self.lq_range = lq_range
            self.dt_range = dt_range

            final_df = dep_events.query(
                f"{lq_param} > {lq_range[0]} & {lq_param} < {lq_range[1]} & {self.dt_param} > {dt_range[0]} & {self.dt_param} < {dt_range[1]}"
            )

            result = linregress(
                final_df[self.dt_param],
                final_df[lq_param],
                alternative="greater",
            )
            self.dt_fit_pars = result

            df["LQ_Corrected"] = (
                df[lq_param]
                - df[self.dt_param] * self.dt_fit_pars[0]
                - self.dt_fit_pars[1]
            )

        except BaseException as e:
            if e == KeyboardInterrupt:
                raise (e)
            elif self.debug_mode:
                raise (e)
            log.error("LQ drift time correction failed")
            self.dt_fit_pars = (np.nan, np.nan)

        self.update_cal_dicts(
            {
                "LQ_Corrected": {
                    "expression": f"{lq_param} - dt_eff*a - b",
                    "parameters": {"a": self.dt_fit_pars[0], "b": self.dt_fit_pars[1]},
                }
            }
        )

    def get_cut_lq_dep(self, df: pd.DataFrame(), lq_param: str, cal_energy_param: str):
        """
        Determines the cut value for LQ. Value is calculated by fitting the LQ distribution
        for events in the DEP to a gaussian. The LQ values are normalized by the fitted
        sigma, and the cut value is set to a value of 3. Sideband subtraction is used to
        determine the LQ distribution for DEP events. Events greater than the cut value
        fail the cut.
        """

        log.info("Starting LQ Cut calculation")
        try:

            pars, errs, hist, bins = binned_lq_fit(
                df, lq_param, cal_energy_param, peak=1592.5
            )

            self.cut_fit_pars = pars
            self.cut_fit_errs = errs
            self.fit_hist = (hist, bins)
            self.cut_val = 3

            df["LQ_Classifier"] = np.divide(df[lq_param].to_numpy(), pars[1])
            df["LQ_Cut"] = df["LQ_Classifier"] < self.cut_val

        except BaseException as e:
            if e == KeyboardInterrupt:
                raise (e)
            elif self.debug_mode:
                raise (e)
            log.error("LQ cut determination failed")
            self.cut_val = np.nan
            c = cost.UnbinnedNLL(np.array([0]), gaussian.pdf)
            m = Minuit(c, np.full(2, np.nan))
            self.cut_fit_pars = pars = m.values

        self.update_cal_dicts(
            {
                "LQ_Classifier": {
                    "expression": f"({lq_param} / a)",
                    "parameters": {"a": pars[1]},
                },
                "LQ_Cut": {
                    "expression": "(LQ_Classifier < a)",
                    "parameters": {"a": self.cut_val},
                },
            }
        )

    def calibrate(self, df, initial_lq_param):
        """Run the LQ calibration and calculate the cut value"""

        self.lq_timecorr(df, initial_lq_param)
        log.info("Finished LQ Time Correction")

        self.drift_time_correction(
            df, lq_param="LQ_Timecorr", cal_energy_param=self.cal_energy_param
        )
        log.info("Finished LQ Drift Time Correction")

        self.get_cut_lq_dep(
            df, lq_param="LQ_Corrected", cal_energy_param=self.cal_energy_param
        )
        log.info("Finished Calculating the LQ Cut Value")

        final_lq_param = "LQ_Classifier"
        peaks_of_interest = [1592.5, 2039, 2103.53, 2614.50]
        self.low_side_sf = pd.DataFrame()
        fit_widths = [(40, 25), (0, 0), (25, 40), (50, 50)]
        self.low_side_peak_dfs = {}

        log.info("Calculating peak survival fractions")
        for i, peak in enumerate(peaks_of_interest):
            try:
                select_df = df.query(f"{self.selection_string}")
                fwhm = self.eres_func(peak)
                if peak == 2039:
                    emin = 2 * fwhm
                    emax = 2 * fwhm
                    peak_df = select_df.query(
                        f"({self.cal_energy_param}>{peak-emin})&({self.cal_energy_param}<{peak+emax})"
                    )

                    cut_df, sf, sf_err = compton_sf_sweep(
                        peak_df[self.cal_energy_param].to_numpy(),
                        peak_df[final_lq_param].to_numpy(),
                        self.cut_val,
                        cut_range=(0, 5),
                        n_samples=30,
                        mode="less",
                    )
                    self.low_side_sf = pd.concat(
                        [
                            self.low_side_sf,
                            pd.DataFrame([{"peak": peak, "sf": sf, "sf_err": sf_err}]),
                        ]
                    )
                    self.low_side_peak_dfs[peak] = cut_df
                else:
                    emin, emax = fit_widths[i]
                    fit_range = (peak - emin, peak + emax)
                    peak_df = select_df.query(
                        f"({self.cal_energy_param}>{fit_range[0]})&({self.cal_energy_param}<{fit_range[1]})"
                    )
                    cut_df, sf, sf_err = get_sf_sweep(
                        peak_df[self.cal_energy_param].to_numpy(),
                        peak_df[final_lq_param].to_numpy(),
                        self.cut_val,
                        peak,
                        fwhm,
                        fit_range=fit_range,
                        cut_range=(0, 5),
                        n_samples=30,
                        mode="less",
                    )
                    self.low_side_sf = pd.concat(
                        [
                            self.low_side_sf,
                            pd.DataFrame([{"peak": peak, "sf": sf, "sf_err": sf_err}]),
                        ]
                    )
                    self.low_side_peak_dfs[peak] = cut_df
                log.info(f"{peak}keV: {sf:2.1f} +/- {sf_err:2.1f} %")
            except BaseException as e:
                if e == KeyboardInterrupt:
                    raise (e)
                elif self.debug_mode:
                    raise (e)
                self.low_side_sf = pd.concat(
                    [
                        self.low_side_sf,
                        pd.DataFrame([{"peak": peak, "sf": np.nan, "sf_err": np.nan}]),
                    ]
                )
                log.error(f"LQ Survival fraction determination failed for {peak} peak")
        self.low_side_sf.set_index("peak", inplace=True)


def plot_lq_mean_time(
    lq_class, data, lq_param="LQ_Timecorr", figsize=(12, 8), fontsize=12
) -> plt.figure:
    """Plots the mean LQ value calculated for each given timestamp"""

    plt.rcParams["figure.figsize"] = figsize
    plt.rcParams["font.size"] = fontsize
    fig, ax = plt.subplots(1, 1)
    try:
        ax.errorbar(
            [
                datetime.strptime(tstamp, "%Y%m%dT%H%M%SZ")
                for tstamp in lq_class.timecorr_df.index
            ],
            lq_class.timecorr_df["mean"],
            yerr=lq_class.timecorr_df["mean_err"],
            linestyle=" ",
        )

        grouped_means = [
            cal_dict["LQ_Timecorr"]["parameters"]["a"]
            for tstamp, cal_dict in lq_class.cal_dicts.items()
        ]
        ax.step(
            [
                datetime.strptime(tstamp, "%Y%m%dT%H%M%SZ")
                for tstamp in lq_class.cal_dicts
            ],
            grouped_means,
            where="post",
        )
        ax.fill_between(
            [
                datetime.strptime(tstamp, "%Y%m%dT%H%M%SZ")
                for tstamp in lq_class.cal_dicts
            ],
            y1=np.array(grouped_means) - 0.2 * np.array(lq_class.timecorr_df["res"]),
            y2=np.array(grouped_means) + 0.2 * np.array(lq_class.timecorr_df["res"]),
            color="green",
            alpha=0.2,
        )
        ax.fill_between(
            [
                datetime.strptime(tstamp, "%Y%m%dT%H%M%SZ")
                for tstamp in lq_class.cal_dicts
            ],
            y1=np.array(grouped_means) - 0.4 * np.array(lq_class.timecorr_df["res"]),
            y2=np.array(grouped_means) + 0.4 * np.array(lq_class.timecorr_df["res"]),
            color="yellow",
            alpha=0.2,
        )
    except Exception:
        pass
    ax.set_xlabel("time")
    ax.set_ylabel("LQ mean")
    myfmt = mdates.DateFormatter("%b %d")
    ax.xaxis.set_major_formatter(myfmt)
    plt.close()
    return fig


def plot_drift_time_correction(
    lq_class, data, lq_param="LQ_Timecorr", figsize=(12, 8), fontsize=12
) -> plt.figure:
    """Plots a 2D histogram of LQ versus effective drift time in a 6 keV
    window around the DEP. Additionally plots the fit results for the
    drift time correction."""

    plt.rcParams["figure.figsize"] = figsize
    plt.rcParams["font.size"] = fontsize
    fig, ax = plt.subplots(1, 1)

    try:
        dep_range = (1590, 1595)

        initial_df = data[
            (data[lq_class.cal_energy_param] > dep_range[0])
            & (data[lq_class.cal_energy_param] < dep_range[1])
        ]
        max_dt = 1500
        max_lq = 2.5

        plt.hist2d(
            initial_df["dt_eff"],
            initial_df[lq_param],
            bins=100,
            range=((0, max_dt), (0, max_lq)),
            norm=mcolors.LogNorm(),
        )

        x = np.linspace(0, max_dt, 100)
        model = lq_class.dt_fit_pars[0] * x + lq_class.dt_fit_pars[1]

        plt.plot(x, model, color="r")

        plt.axvline(lq_class.dt_range[0], color="k")
        plt.axvline(lq_class.dt_range[1], color="k")
        plt.axhline(lq_class.lq_range[0], color="k")
        plt.axhline(lq_class.lq_range[1], color="k")

        plt.xlabel("Drift Time (ns)")
        plt.ylabel("LQ")

        plt.title("LQ versus Drift Time for DEP")

    except Exception:
        pass

    plt.tight_layout()
    plt.close()
    return fig


def plot_lq_cut_fit(lq_class, data, figsize=(12, 8), fontsize=12) -> plt.figure:
    """Plots the final histogram of LQ values for events in the
    DEP, and the fit results used for determining the cut
    value"""

    plt.rcParams["figure.figsize"] = figsize
    plt.rcParams["font.size"] = fontsize
    fig, (ax1, ax2) = plt.subplots(2, 1)
    try:
        hist, bins = lq_class.fit_hist
        fit_pars = lq_class.cut_fit_pars

        x_low = bins[0]
        x_high = bins[-1]

        ax1.stairs(hist, bins, label="data")
        xs = np.linspace(round(bins[0], 3), round(bins[-1], 3), len(bins) - 1)
        ls = np.sum(hist)
        dx = np.diff(bins)
        ax1.plot(
            xs,
            gaussian.pdf_norm(xs, x_low, x_high, fit_pars[0], fit_pars[1]) * dx * ls,
            label="Gaussian Fit",
        )

        # ax1.set_xlabel('LQ')
        ax1.set_title("Fit of LQ events in DEP")
        ax1.legend()

        bin_centers = (bins[:-1] + bins[1:]) / 2
        reses = (
            hist
            - (
                gaussian.pdf_norm(bin_centers, x_low, x_high, fit_pars[0], fit_pars[1])
                * dx
                * ls
            )
        ) / (
            gaussian.pdf_norm(bin_centers, x_low, x_high, fit_pars[0], fit_pars[1])
            * dx
            * ls
        )
        ax2.plot(bin_centers, reses, marker="s", linestyle="")
        ax2.set_xlabel("LQ")
        ax2.set_ylabel("residuals")

    except Exception:
        pass

    plt.tight_layout()
    plt.close()
    return fig


def plot_survival_fraction_curves(
    lq_class, data, figsize=(12, 8), fontsize=12
) -> plt.figure:
    """Plots the survival fraction curves as a function of
    LQ cut values for every peak of interest"""

    plt.rcParams["figure.figsize"] = figsize
    plt.rcParams["font.size"] = fontsize

    fig = plt.figure()
    try:
        plt.vlines(
            lq_class.cut_val,
            0,
            100,
            label=f"cut value: {lq_class.cut_val:1.2f}",
            color="black",
        )

        for peak, survival_df in lq_class.low_side_peak_dfs.items():
            try:
                plt.errorbar(
                    survival_df.index,
                    survival_df["sf"],
                    yerr=survival_df["sf_err"],
                    label=f'{AoE.get_peak_label(peak)} {peak} keV: {lq_class.low_side_sf.loc[peak]["sf"]:2.1f} +/- {lq_class.low_side_sf.loc[peak]["sf_err"]:2.1f} %',
                )
            except Exception:
                pass
    except Exception:
        pass
    vals, labels = plt.yticks()
    plt.yticks(vals, [f"{x:,.0f} %" for x in vals])
    plt.legend(loc="lower right")
    plt.xlabel("cut value")
    plt.ylabel("survival percentage")
    plt.ylim([0, 105])
    plt.close()
    return fig


def plot_sf_vs_energy(
    lq_class, data, xrange=(900, 3000), n_bins=701, figsize=(12, 8), fontsize=12
) -> plt.figure:
    """Plots the survival fraction as a function of energy"""

    plt.rcParams["figure.figsize"] = figsize
    plt.rcParams["font.size"] = fontsize

    fig = plt.figure()
    try:
        bins = np.linspace(xrange[0], xrange[1], n_bins)
        counts_pass, bins_pass, _ = pgh.get_hist(
            data.query(f"{lq_class.selection_string}&LQ_Cut")[
                lq_class.cal_energy_param
            ],
            bins=bins,
        )
        counts, bins, _ = pgh.get_hist(
            data.query(lq_class.selection_string)[lq_class.cal_energy_param],
            bins=bins,
        )
        survival_fracs = counts_pass / (counts + 10**-99)

        plt.step(pgh.get_bin_centers(bins_pass), 100 * survival_fracs)
    except Exception:
        pass
    plt.ylim([0, 100])
    vals, labels = plt.yticks()
    plt.yticks(vals, [f"{x:,.0f} %" for x in vals])
    plt.xlabel("energy (keV)")
    plt.ylabel("survival percentage")
    plt.close()
    return fig


def plot_spectra(
    lq_class,
    data,
    xrange=(900, 3000),
    n_bins=2101,
    xrange_inset=(1580, 1640),
    n_bins_inset=200,
    figsize=(12, 8),
    fontsize=12,
) -> plt.figure:
    """Plots a 2D histogram of the LQ classifier vs calibrated energy"""

    plt.rcParams["figure.figsize"] = figsize
    plt.rcParams["font.size"] = fontsize

    fig, ax = plt.subplots()
    try:
        bins = np.linspace(xrange[0], xrange[1], n_bins)
        ax.hist(
            data.query(lq_class.selection_string)[lq_class.cal_energy_param],
            bins=bins,
            histtype="step",
            label="before PSD",
        )
        # ax.hist(
        #     data.query(f"{lq_class.selection_string}&AoE_Double_Sided_Cut")[
        #         lq_class.cal_energy_param
        #     ],
        #     bins=bins,
        #     histtype="step",
        #     label="after double sided A/E cut",
        # )
        ax.hist(
            data.query(f"{lq_class.selection_string}&LQ_Cut")[
                lq_class.cal_energy_param
            ],
            bins=bins,
            histtype="step",
            label="after LQ cut",
        )
        ax.hist(
            data.query(f"{lq_class.selection_string} & (~LQ_Cut)")[
                lq_class.cal_energy_param
            ],
            bins=bins,
            histtype="step",
            label="rejected by LQ cut",
        )

        axins = ax.inset_axes([0.25, 0.07, 0.4, 0.3])
        bins = np.linspace(xrange_inset[0], xrange_inset[1], n_bins_inset)
        select_df = data.query(
            f"{lq_class.cal_energy_param}<{xrange_inset[1]}&{lq_class.cal_energy_param}>{xrange_inset[0]}"
        )
        axins.hist(
            select_df.query(lq_class.selection_string)[lq_class.cal_energy_param],
            bins=bins,
            histtype="step",
        )
        # axins.hist(
        #     select_df.query(f"{lq_class.selection_string}&AoE_Double_Sided_Cut")[
        #         lq_class.cal_energy_param
        #     ],
        #     bins=bins,
        #     histtype="step",
        # )
        axins.hist(
            select_df.query(f"{lq_class.selection_string}&LQ_Cut")[
                lq_class.cal_energy_param
            ],
            bins=bins,
            histtype="step",
        )
        axins.hist(
            select_df.query(f"{lq_class.selection_string} & (~LQ_Cut)")[
                lq_class.cal_energy_param
            ],
            bins=bins,
            histtype="step",
        )
    except Exception:
        pass
    ax.set_xlim(xrange)
    ax.set_yscale("log")
    plt.xlabel("energy (keV)")
    plt.ylabel("counts")
    plt.legend(loc="upper left")
    plt.close()
    return fig


def plot_classifier(
    lq_class,
    data,
    lq_param="LQ_Classifier",
    xrange=(800, 3000),
    yrange=(-10, 30),
    xn_bins=700,
    yn_bins=500,
    figsize=(12, 8),
    fontsize=12,
) -> plt.figure:
    plt.rcParams["figure.figsize"] = figsize
    plt.rcParams["font.size"] = fontsize

    fig = plt.figure()
    try:
        plt.hist2d(
            data.query(lq_class.selection_string)[lq_class.cal_energy_param],
            data.query(lq_class.selection_string)[lq_param],
            bins=[
                np.linspace(xrange[0], xrange[1], xn_bins),
                np.linspace(yrange[0], yrange[1], yn_bins),
            ],
            norm=LogNorm(),
        )
    except Exception:
        pass
    plt.xlabel("energy (keV)")
    plt.ylabel(lq_param)
    plt.xlim(xrange)
    plt.ylim(yrange)
    plt.close()
    return fig



r"""
A class that creates the sum of distributions, with methods for scipy computed :func:`pdfs` and :func:`cdfs`, as well as fast :func:`get_pdfs`, and :func:`pdf_ext`

.. code-block:: python

    mu1, mu2, sigma, frac = range(4)
    moyal_add = SumDists([(moyal, [mu1, sigma]), (moyal, [mu2, sigma])], [frac], "fracs") # create two moyals that share a sigma and differ by a fraction,
    x = np.arange(-10,10)
    pars = np.array([1, 2, 2, 0.1]) # corresponds to mu1 = 1, mu2 = 2, sigma = 2, frac=0.1
    moyal_add.pdf(x, *pars)
    moyal_add.draw_pdf(x, *pars)
    moyal_add.required_args()
"""

import inspect
from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np
from scipy.stats._distn_infrastructure import rv_continuous

from pygama.math.functions.pygama_continuous import PygamaContinuous


def get_dists_and_par_idxs(
    dists_and_pars_array: np.array(tuple, tuple),
) -> Tuple[np.array, np.array]:
    r"""
    Split the array of tuples passed to the :func:`SumDists` constructor into separate arrays with one containing only the
    distributions and the other containing the parameter index arrays. Also performs some sanity checks.

    Parameters
    ----------
    dists_and_pars_array
        An array of tuples, each tuple contains a distribution and the indices of the parameters
        in the :func:`SumDists` call that correspond to that distribution

    Returns
    -------
    dists, par_idxs
        Returned only if every distribution has the correct number of required arguments
    """
    # can only sum two dists at once, check:
    if len(dists_and_pars_array) != 2:
        raise ValueError("Can only sum two distributions at once.")

    dists = []
    par_idxs = []

    # Check that each tuple in the dists_and_pars_array is of length two
    for dist_and_pars in dists_and_pars_array:
        if len(dist_and_pars) != 2:
            raise ValueError(
                "Each tuple needs a distribution and a parameter index array."
            )
        else:
            dists.append(dist_and_pars[0])
            par_idxs.append(np.array(dist_and_pars[1]))

    return dists, par_idxs


def get_areas_fracs(
    params: np.array,
    area_frac_idxs: np.array,
    frac_flag: bool,
    area_flag: bool,
    one_area_flag: bool,
) -> Tuple[np.array, np.array]:
    r"""
    Grab the value(s) of either the fraction or the areas passed in the params array from the :func:`SumDists` call.
    If :func:`SumDists` is in "fracs" mode, then this grabs `f` from the params array and returns fracs = [f, 1-f] and areas of unity.
    If :func:`SumDists` is in "areas" mode, then this grabs `s, b` from the params array and returns unity fracs and areas = [s, b]
    If :func:`SumDists` is in "one_area" mode, then this grabs `s` from the params array and returns unity fracs and areas = [s, 1]

    Parameters
    ----------
    params
        An array containing the shape values from a :func:`SumDists` call
    area_frac_idxs
        An array containing the indices of either the fracs or the areas present in the params array
    frac_flag
        A boolean telling if :func:`SumDists` is in fracs mode or not
    area_flag
        A boolean telling if :func:`SumDists` is in areas mode or not
    one_area_flag
        A boolean telling if :func:`SumDists` is to apply only area to one distribution

    Returns
    -------
    fracs, areas
        Values of the fractions and the areas to post-multiply the sum of the distributions with
    """
    if frac_flag:
        fracs = np.array([params[area_frac_idxs[0]], 1 - params[area_frac_idxs[0]]])
        areas = np.array([1, 1])
    elif area_flag:
        fracs = np.array([1, 1])
        areas = np.array([*params[area_frac_idxs]])
    elif one_area_flag:
        fracs = np.array([1, 1])
        areas = np.array([*params[area_frac_idxs], 1])

    else:
        fracs = np.array([1, 1])
        areas = np.array([1, 1])

    return fracs, areas


def get_parameter_names(dists: np.array, par_idxs: np.array, par_size: int) -> np.array:
    r"""
    Returns an array of the names of the required parameters for an instance of :func:`SumDists`
    Works by calling :func:`.required_args` for each distribution present in the sum.
    If a parameter is shared between distributions its name is only added once.
    If two parameters are required and share a name, then the second parameter gets an added index at the end.

    Parameters
    ----------
    dists
        An array containing the distributions in this instance of :func:`SumDists`
    par_idxs
        An array of arrays, each array contains the indices of the parameters in the :func:`SumDists` call that correspond to that distribution
    par_size
        The size of the single parameter index array

    Returns
    -------
    param_names
        An array containing the required parameter names
    """
    param_names = np.empty(par_size + 1, dtype=object)
    overall_par_idxs = []
    for i in range(len(dists)):
        mask = ~np.isin(
            par_idxs[i], overall_par_idxs
        )  # see if indices of the required args that are not included yet
        new_idxs = par_idxs[i][
            mask
        ]  # get indices of the required args that are not included yet
        prereq_names = np.array(dists[i].required_args())[mask]
        req_names = []
        # Check for duplicate names
        for name in prereq_names:
            if name in param_names:
                if name[-1].isdigit():
                    name = name[:-1] + f"{int(name[-1])+1}"
                else:
                    name = name + "1"
            req_names.append(name)

        param_names[new_idxs] = req_names
        overall_par_idxs.extend(par_idxs[i])
    return param_names


def copy_signature(signature_to_copy, obj_to_copy_to):
    """
    Copy the signature provided in signature_to_copy into the signature for "obj_to_copy_to".
    This is necessary so that we can override the signature for the various methods attached to
    different objects.
    """

    def wrapper(*args, **kwargs):
        return obj_to_copy_to(*args, **kwargs)

    wrapper.__signature__ = signature_to_copy
    return wrapper


class SumDists(rv_continuous):
    r"""
    Initialize an rv_continuous method so that we gain access to scipy computable methods.
    Precompute the support of the sum of the distributions.

    The correct way to initialize is SumDists([(d1, [p1]), (d2, [p2])], [area_idx_1/frac_idx, area_idx_2/], frac_flag)
    Where d_i is a distribution and p_i is a parameter index array for that distribution.

    Parameter index arrays contain indices that slice a single parameter array that is passed to method calls.
    For example, if the user will eventually pass parameters=[mu, sigma, tau, frac] to function.get_pdf(x, parameters)
    and the first distribution takes (mu, sigma) as its parameters, then p1=[0,1]. If the second distribution takes (mu, sigma, tau)
    then its parameter index array would be p2=[0, 1, 2] because tau is the index 2 entry in parameters.

    Each par array can contain [x_lo, x_hi, mu, sigma, shape], and *must be placed in that order*.

    The single parameter array passed to function calls *must* follow the ordering convention [x_lo, x_hi, frac/areas, shapes_1, frac/areas2, shapes_2]
    The single parameter array that is passed to :func:`pdf_norm`, :func:`pdf_ext`, and :func:`cdf_norm` calls *must* have x_lo, x_hi
    as its first two arguments if none of the distributions require an explicit definition of their support.

    There are 4 flag options:

    1. flag = "areas", two areas passed as fit variables
    2. flag = "fracs", one fraction is passed a a fit variable, with the convention that SumDists performs f*dist_1 + (1-f)*dist_2
    3. flag = "one_area", one area is passed a a fit variable, with the convention that SumDists performs area*dist_1 + 1*dist_2
    4. flag = None, no areas or fracs passed as fit variables, both are normalized to unity.

    Notes
    -----
    dists must be unfrozen pygama distributions of the type :func:`PygamaContinuous` or :func:`sum_dist`.
    """

    def _argcheck(self, *args) -> bool:
        """
        Check that each distribution gets its own valid arguments.

        Notes
        -----
        This overloads the :func:`scipy` definition so that the methods :func:`.pdf` work.
        """
        args = np.array(args)
        cond = True
        for i, dist in enumerate(self.dists):
            cond = np.logical_and(cond, dist._argcheck(*args[self.par_idxs[i]]))
        return cond

    def __init__(
        self,
        dists_and_pars_array,
        area_frac_idxs,
        flag,
        parameter_names=None,
        components=False,
        support_required=False,
        **kwds,
    ):
        """
        Parameters
        ----------
        dists_and_pars_array
            A list of two tuples, containing [(dist_1, [mu1, sigma1, shapes1]), (dist_2, [mu2, sigma2, shapes2])] to create a sum_dist from
        area_frac_idxs
            A list of the indices at which that either the areas or fraction will be placed in the eventual method calls
        flag
            One of three strings that initialize :func:`sum_dist` in different modes. Either "fracs", "areas" or "one_area".
        parameter_names
            An optional list of strings that contain the parameters names in the order they will appear in method calls
        components
            A boolean that if true will cause methods to return components instead of the sum of the distributions
        support_required
            A boolean that if true tells :func:`sum_dist` that x_lo, x_hi will *always* be passed in method calls.
        """
        # Extract the distributions and parameter index arrays from the constructor
        dists, par_idxs = get_dists_and_par_idxs(dists_and_pars_array)

        self.dists = dists
        self.par_idxs = par_idxs
        self.area_frac_idxs = area_frac_idxs
        self.components = components
        self.support_required = support_required

        # Check that the dists are in fact distributions
        for i in range(len(dists)):
            if (not isinstance(dists[i], PygamaContinuous)) and (
                not isinstance(dists[i], SumDists)
            ):
                raise ValueError(
                    f"Distribution at index {i} has value {dists[i]},\
                and is an array and not a PygamaContinuous distribution"
                )

        # Get the parameter names for later introspection
        # First, find the length of the eventual single parameter array
        par_size = 0
        for par_idx in par_idxs:
            if np.amax(par_idx) >= par_size:
                par_size = np.amax(par_idx)
        par_size = (
            np.amax([par_size, np.amax(area_frac_idxs)])
            if len(area_frac_idxs) != 0
            else par_size
        )
        shapes = get_parameter_names(dists, par_idxs, par_size)

        # Set the internal state depending on what flag was passed
        if flag == "fracs":
            if len(area_frac_idxs) != 1:
                raise ValueError(
                    "SumDists only accepts the parameter position of one fraction."
                )
            self.frac_flag = True
            self.area_flag = False
            self.one_area_flag = False
            shapes[area_frac_idxs[0]] = "f"  # add frac name to the shapes
        elif flag == "areas":
            if len(area_frac_idxs) != 2:
                raise ValueError("SumDists needs two parameter indices of areas.")
            self.frac_flag = False
            self.area_flag = True
            self.one_area_flag = False
            shapes[area_frac_idxs] = ["s", "b"]  # add area names to the shapes
        elif flag == "one_area":
            # needed so that we can create SumDists from SumDists without having an overall area to the second dist
            # Sets the area of the second dist to 1
            if len(area_frac_idxs) != 1:
                raise ValueError("SumDists needs one parameter index of an area.")
            self.frac_flag = False
            self.area_flag = False
            self.one_area_flag = True
            shapes[area_frac_idxs] = "s"  # add area name to the shapes
        else:
            self.frac_flag = False
            self.area_flag = False
            self.one_area_flag = False

        shapes = list(shapes)
        # override the parameter names if string is passed to the constructor
        if parameter_names is not None:
            shapes = parameter_names

        # If a support is explicitly passed, flag it.
        # Record the index of x_lo and x_hi if they are required, if they aren't set them to 0, 1
        # This is so that :func:`pdf_ext`, :func:`pdf_norm` and :func:`cdf_norm` can find the x_lo, x_hi if support is required, otherwise they take x_lo, x_hi from the start of the parameter array
        if "x_lo" in shapes or "x_hi" in shapes:
            self.support_required = True
            self.x_lo_idx = shapes.index("x_lo")
            self.x_hi_idx = shapes.index("x_hi")
            extended_shapes = [
                "x",
                *shapes,
            ]  # get the parameter names for methods that require the fit range passed, like extended fits
            self.extended_shapes = extended_shapes
        else:
            self.x_lo_idx = 0
            self.x_hi_idx = 1
            extended_shapes = [
                "x",
                "x_lo",
                "x_hi",
                *shapes,
            ]  # get the parameter names for methods that require the fit range passed, like extended fits
            self.extended_shapes = extended_shapes

        # Now that we have inferred or set the shapes, store them so that :func:`required_args` can return them
        self.req_args = shapes

        # set attributes for the methods so that Iminuit can introspect parameter names
        x_shapes = [
            "x",
            *shapes,
        ]  # add the x values to the parameter names that are being passed, Iminuit assumes that this is in the first position
        self.x_shapes = x_shapes

        # Scipy requires the argument names as one string with commas in between each parameter name
        shapes = ",".join(str(x) for x in shapes)

        super().__init__(self, shapes=shapes, **kwds)

    def _pdf(self, x, *params):
        """
        Overload :func:`rv_continuous` definition of pdf in order to access other methods.
        """
        pdfs = self.dists
        params = np.array(params)[:, 0]  # scipy ravels the parameter array...

        fracs, areas = get_areas_fracs(
            params,
            self.area_frac_idxs,
            self.frac_flag,
            self.area_flag,
            self.one_area_flag,
        )

        if self.components:
            return areas[0] * fracs[0] * pdfs[0].pdf(
                x, *params[self.par_idxs[0]]
            ), areas[1] * fracs[1] * pdfs[1].pdf(x, *params[self.par_idxs[1]])

        else:
            # This is faster than list comprehension
            probs = areas[0] * fracs[0] * pdfs[0].pdf(
                x, *params[self.par_idxs[0]]
            ) + areas[1] * fracs[1] * pdfs[1].pdf(x, *params[self.par_idxs[1]])
            return probs

    def _cdf(self, x, *params):
        """
        Overload :func:`rv_continuous` definition of cdf in order to access other methods.
        """
        cdfs = self.dists
        params = np.array(params)[:, 0]

        fracs, areas = get_areas_fracs(
            params,
            self.area_frac_idxs,
            self.frac_flag,
            self.area_flag,
            self.one_area_flag,
        )

        if self.components:
            return areas[0] * fracs[0] * cdfs[0].cdf(
                x, *params[self.par_idxs[0]]
            ), areas[1] * fracs[1] * cdfs[1].cdf(x, *params[self.par_idxs[1]])

        else:
            # This is faster than list comprehension
            probs = areas[0] * fracs[0] * cdfs[0].cdf(
                x, *params[self.par_idxs[0]]
            ) + areas[1] * fracs[1] * cdfs[1].cdf(x, *params[self.par_idxs[1]])
            return probs

    def get_pdf(self, x, *params):
        """
        Returns the specified sum of all distributions' :func:`get_pdf` methods.
        """
        pdfs = self.dists
        params = np.array(params)

        fracs, areas = get_areas_fracs(
            params,
            self.area_frac_idxs,
            self.frac_flag,
            self.area_flag,
            self.one_area_flag,
        )

        if self.components:
            return areas[0] * fracs[0] * pdfs[0].get_pdf(
                x, *params[self.par_idxs[0]]
            ), areas[1] * fracs[1] * pdfs[1].get_pdf(x, *params[self.par_idxs[1]])

        else:
            # This is faster than list comprehension
            probs = areas[0] * fracs[0] * pdfs[0].get_pdf(
                x, *params[self.par_idxs[0]]
            ) + areas[1] * fracs[1] * pdfs[1].get_pdf(x, *params[self.par_idxs[1]])
            return probs

    def get_cdf(self, x, *params):
        """
        Returns the specified sum of all distributions' :func:`get_cdf` methods.
        """
        cdfs = self.dists
        params = np.array(params)

        fracs, areas = get_areas_fracs(
            params,
            self.area_frac_idxs,
            self.frac_flag,
            self.area_flag,
            self.one_area_flag,
        )

        if self.components:
            return areas[0] * fracs[0] * cdfs[0].get_cdf(
                x, *params[self.par_idxs[0]]
            ), areas[1] * fracs[1] * cdfs[1].get_cdf(x, *params[self.par_idxs[1]])

        else:
            # This is faster than list comprehension
            probs = areas[0] * fracs[0] * cdfs[0].get_cdf(
                x, *params[self.par_idxs[0]]
            ) + areas[1] * fracs[1] * cdfs[1].get_cdf(x, *params[self.par_idxs[1]])
            return probs

    def pdf_norm(self, x, *params):
        """
        Returns the specified sum of all distributions' :func:`get_pdf` methods, but normalized on a range x_lo to x_hi. Used in unbinned NLL fits.

        NOTE: This assumes that x_lo, x_hi are the first two elements in the parameter array, unless x_lo and x_hi are required to define the support
        of the :func:`SumDists` for *all* method calls. For :func:`SumDists` created from a distribution where the support needs
        to be explicitly passed, this means that :func:`pdf_norm` sets the fit range equal to the support range.
        It also means that summing distributions with two different explicit supports is not allowed, e.g. we cannot sum two step functions with different supports.
        """
        # Grab the values of x_lo and x_hi from the parameters. For functions where the support needs to be defined, this can be anywhere in the parameter array
        # Otherwise, x_lo and x_hi *must* be the first two values passed.
        x_lo = params[self.x_lo_idx]
        x_hi = params[self.x_hi_idx]
        if not self.support_required:
            params = params[
                2:
            ]  # chop off x_lo, x_hi from the shapes to actually pass to get_cdf

        norm = np.diff(self.get_cdf(np.array([x_lo, x_hi]), *params))

        if norm == 0:
            return np.full_like(x, np.inf)
        else:
            return self.get_pdf(x, *params) / norm

    def cdf_norm(self, x, *params):
        """
        Returns the specified sum of all distributions' :func:`get_cdf` methods, but normalized on a range x_lo to x_hi. Used in binned NLL fits.

        NOTE: This assumes that x_lo, x_hi are the first two elements in the parameter array, unless x_lo and x_hi are required to define the support
        of the :func:`SumDists` for *all* method calls. For :func:`SumDists` created from a distribution where the support needs
        to be explicitly passed, this means that :func:`cdf_norm` sets the fit range equal to the support range.
        It also means that summing distributions with two different explicit supports is not allowed, e.g. we cannot sum two step functions with different supports.
        """
        # Grab the values of x_lo and x_hi from the parameters. For functions where the support needs to be defined, this can be anywhere in the parameter array
        # Otherwise, x_lo and x_hi *must* be the first two values passed.
        x_lo = params[self.x_lo_idx]
        x_hi = params[self.x_hi_idx]
        if not self.support_required:
            params = params[
                2:
            ]  # chop off x_lo, x_hi from the shapes to actually pass to get_cdf

        norm = np.diff(self.get_cdf(np.array([x_lo, x_hi]), *params))

        if norm == 0:
            return np.full_like(x, np.inf)
        else:
            return (self.get_cdf(x, *params)) / norm

    def pdf_ext(self, x, *params):
        """
        Returns the specified sum of all distributions' :func:`pdf_ext` methods, normalized on a range x_lo to x_hi. Used in extended unbinned NLL fits.

        NOTE: This assumes that x_lo, x_hi are the first two elements in the parameter array, unless x_lo and x_hi are required to define the support
        of the :func:`SumDists` for *all* method calls. For :func:`SumDists` created from a distribution where the support needs
        to be explicitly passed, this means that :func:`pdf_ext` sets the fit range equal to the support range.
        It also means that summing distributions with two different explicit supports is not allowed, e.g. we cannot sum two step functions with different supports.
        """
        pdf_exts = self.dists
        params = np.array(params)

        # Grab the values of x_lo and x_hi from the parameters. For functions where the support needs to be defined, this can be anywhere in the parameter array
        # Otherwise, x_lo and x_hi *must* be the first two values passed.
        x_lo = params[self.x_lo_idx]
        x_hi = params[self.x_hi_idx]
        if not self.support_required:
            params = params[
                2:
            ]  # chop off x_lo, x_hi from the shapes to actually pass to get_cdf

        fracs, areas = get_areas_fracs(
            params,
            self.area_frac_idxs,
            self.frac_flag,
            self.area_flag,
            self.one_area_flag,
        )

        # sig = areas[0] + areas[1] # this is a hack, it performs faster but may not *always* be true
        sig = (
            areas[0]
            * fracs[0]
            * np.diff(
                pdf_exts[0].get_cdf(np.array([x_lo, x_hi]), *params[self.par_idxs[0]])
            )[0]
            + areas[1]
            * fracs[1]
            * np.diff(
                pdf_exts[1].get_cdf(np.array([x_lo, x_hi]), *params[self.par_idxs[1]])
            )[0]
        )

        if self.components:
            return (
                sig,
                areas[0] * fracs[0] * pdf_exts[0].get_pdf(x, *params[self.par_idxs[0]]),
                areas[1] * fracs[1] * pdf_exts[1].get_pdf(x, *params[self.par_idxs[1]]),
            )
        else:
            probs = areas[0] * fracs[0] * pdf_exts[0].get_pdf(
                x, *params[self.par_idxs[0]]
            ) + areas[1] * fracs[1] * pdf_exts[1].get_pdf(x, *params[self.par_idxs[1]])

            return sig, probs

    def cdf_ext(self, x, *params):
        """
        Returns the specified sum of all distributions' :func:`get_cdf` methods, used in extended binned NLL fits.
        """
        cdf_exts = self.dists
        params = np.array(params)

        fracs, areas = get_areas_fracs(
            params,
            self.area_frac_idxs,
            self.frac_flag,
            self.area_flag,
            self.one_area_flag,
        )

        if self.components:
            return areas[0] * fracs[0] * cdf_exts[0].get_cdf(
                x, *params[self.par_idxs[0]]
            ), areas[1] * fracs[1] * cdf_exts[1].get_cdf(x, *params[self.par_idxs[1]])

        else:
            # This is faster than list comprehension
            probs = areas[0] * fracs[0] * cdf_exts[0].get_cdf(
                x, *params[self.par_idxs[0]]
            ) + areas[1] * fracs[1] * cdf_exts[1].get_cdf(x, *params[self.par_idxs[1]])
            return probs

    def required_args(self) -> list:
        """
        Returns
        -------
        req_args
            A list of the required arguments for the :func:`SumDists` instance, either passed by the user, or inferred.
        """
        return self.req_args

    def draw_pdf(self, x: np.ndarray, *params) -> None:
        plt.plot(x, self.get_pdf(x, *params))

    def draw_cdf(self, x: np.ndarray, *params) -> None:
        plt.plot(x, self.get_cdf(x, *params))

    # Create some convenience functions for fitting

    def get_mu(
        self, pars: np.ndarray, cov: np.ndarray = None, errors: np.ndarray = None
    ) -> tuple:
        r"""
        Get the mu value from the output of a fit quickly

        Parameters
        ----------
        pars
            Array of fit parameters
        cov
            Array of covariances
        errors
            Array of errors

        Returns
        -------
        mu, error
            where mu is the fit value, and error is either from the covariance matrix or directly passed
        """

        req_args = np.array(self.required_args())
        mu_idx = np.where(req_args == "mu")[0][0]
        mu = pars[mu_idx]

        if errors is not None:
            return mu, errors[mu_idx]
        elif cov is not None:
            return mu, np.sqrt(cov[mu_idx][mu_idx])
        else:
            return mu

    def get_mode(
        self, pars: np.ndarray, cov: np.ndarray = None, errors: np.ndarray = None
    ) -> tuple:
        r"""
        Get the mode value from the output of a fit quickly
        Need to overload this to use hpge_peak_fwhm (to avoid a circular import) for when self is an hpge peak

        Parameters
        ----------
        pars
            Array of fit parameters
        cov
            Array of covariances
        errors
            Array of errors

        Returns
        -------
        mu, error
            where mu is the fit value, and error is either from the covariance matrix or directly passed
        """

        req_args = np.array(self.required_args())
        mu_idx = np.where(req_args == "mu")[0][0]
        mu = pars[mu_idx]

        if errors is not None:
            return mu, errors[mu_idx]
        elif cov is not None:
            return mu, np.sqrt(cov[mu_idx][mu_idx])
        else:
            return mu

    def get_fwhm(self, pars: np.ndarray, cov: np.ndarray = None) -> tuple:
        r"""
        Get the fwhm value from the output of a fit quickly
        Need to overload this to use hpge_peak_fwhm (to avoid a circular import) for when self is an hpge peak,
        and otherwise returns 2sqrt(2log(2))*sigma

        Parameters
        ----------
        pars
            Array of fit parameters
        cov
            Optional, array of covariances for calculating error on the fwhm


        Returns
        -------
        fwhm, error
            the value of the fwhm and its error
        """

        req_args = np.array(self.required_args())
        sigma_idx = np.where(req_args == "sigma")[0][0]

        if cov is None:
            return pars[sigma_idx] * 2 * np.sqrt(2 * np.log(2))
        else:
            return pars[sigma_idx] * 2 * np.sqrt(2 * np.log(2)), np.sqrt(
                cov[sigma_idx][sigma_idx]
            ) * 2 * np.sqrt(2 * np.log(2))

    def get_fwfm(self, pars: np.ndarray, cov: np.ndarray = None, frac_max=0.5) -> tuple:
        r"""
        Get the fwfm value from the output of a fit quickly
        Need to overload this to use hpge_peak_fwfm (to avoid a circular import) for when self is an hpge peak,
        and otherwise returns 2sqrt(2log(2))*sigma

        Parameters
        ----------
        pars
            Array of fit parameters
        cov
            Optional, array of covariances for calculating error on the fwhm


        Returns
        -------
        fwhm, error
            the value of the fwhm and its error
        """

        req_args = np.array(self.required_args())
        sigma_idx = np.where(req_args == "sigma")[0][0]

        if cov is None:
            return pars[sigma_idx] * 2 * np.sqrt(-2 * np.log(frac_max))
        else:
            return pars[sigma_idx] * 2 * np.sqrt(-2 * np.log(frac_max)), np.sqrt(
                cov[sigma_idx][sigma_idx]
            ) * 2 * np.sqrt(-2 * np.log(frac_max))

    def get_total_events(
        self, pars: np.ndarray, cov: np.ndarray = None, errors: np.ndarray = None
    ) -> tuple:
        r"""
        Get the total events from the output of an extended fit quickly
        The total number of events come from the sum of the area of the signal and the area of the background components

        Parameters
        ----------
        pars
            Array of fit parameters
        cov
            Array of covariances
        errors
            Array of errors

        Returns
        -------
        n_total, error
            the total number of events in a spectrum and its associated error
        """

        req_args = np.array(self.required_args())
        n_sig_idx = np.where(req_args == "n_sig")[0][0]
        n_bkg_idx = np.where(req_args == "n_bkg")[0][0]

        if errors is not None:
            return pars[n_sig_idx] + pars[n_bkg_idx], np.sqrt(
                errors[n_sig_idx] ** 2 + errors[n_bkg_idx] ** 2
            )
        elif cov is not None:
            return pars[n_sig_idx] + pars[n_bkg_idx], np.sqrt(
                cov[n_sig_idx][n_sig_idx] ** 2 + cov[n_bkg_idx][n_bkg_idx] ** 2
            )
        else:
            return pars[n_sig_idx] + pars[n_bkg_idx]

    def __getattribute__(self, attr):
        """
        Necessary to overload this so that Iminuit can use inspect.signature to get the correct parameter names
        """
        value = object.__getattribute__(self, attr)

        if attr in ["get_pdf", "get_cdf", "cdf_ext"]:
            params = [
                inspect.Parameter(param, inspect.Parameter.POSITIONAL_OR_KEYWORD)
                for param in self.x_shapes
            ]
            value = copy_signature(inspect.Signature(params), value)
        if attr in [
            "pdf_norm",
            "pdf_ext",
            "cdf_norm",
        ]:  # Set these to include the x_lo, x_hi at the correct positions since these always need those params
            params = [
                inspect.Parameter(param, inspect.Parameter.POSITIONAL_OR_KEYWORD)
                for param in self.extended_shapes
            ]
            value = copy_signature(inspect.Signature(params), value)
        return value



import numpy as np
from scipy.stats import exponnorm, norm

from pygama.math.functions.gauss_on_exgauss import gauss_on_exgauss
from pygama.math.functions.sum_dists import SumDists


def test_gauss_on_exgauss_pdf():
    x = np.arange(-10, 10)
    mu = 1
    sigma = 1
    tau = 3
    h_tail = 0.25

    pars = np.array([mu, sigma, h_tail, tau], dtype=float)

    assert isinstance(gauss_on_exgauss, SumDists)

    y_direct = gauss_on_exgauss.get_pdf(x, *pars)
    scipy_exgauss = h_tail * exponnorm.pdf(
        -1 * x, tau / sigma, -1 * mu, sigma
    )  # to be equivalent to the scipy version, x -> -x, mu -> -mu, k -> k/sigma
    scipy_gauss = (1 - h_tail) * norm.pdf(x, mu, sigma)

    scipy_y = scipy_exgauss + scipy_gauss

    assert np.allclose(y_direct, scipy_y, rtol=1e-8)

    x_lo = -100
    x_hi = 100
    pars = np.array([x_lo, x_hi, mu, sigma, h_tail, tau], dtype=float)

    y_sig, y_ext = gauss_on_exgauss.pdf_ext(x, *pars)
    assert np.allclose(y_ext, scipy_y, rtol=1e-8)
    assert np.allclose(y_sig, 1, rtol=1e-8)


def test_gauss_on_exgauss_cdf():
    x = np.arange(-10, 10)
    mu = 1
    sigma = 1
    tau = 3

    h_tail = 0.25

    pars = np.array([mu, sigma, h_tail, tau], dtype=float)

    assert isinstance(gauss_on_exgauss, SumDists)

    y_direct = gauss_on_exgauss.get_cdf(x, *pars)

    scipy_exgauss = h_tail * (
        1 - exponnorm.cdf(-1 * x, tau / sigma, -1 * mu, sigma)
    )  # to be equivalent to the scipy version, x -> -x, mu -> -mu, k -> k/sigma
    scipy_gauss = (1 - h_tail) * norm.cdf(x, mu, sigma)

    scipy_y = scipy_exgauss + scipy_gauss

    assert np.allclose(y_direct, scipy_y, rtol=1e-8)

    y_ext = gauss_on_exgauss.cdf_ext(x, *pars)
    assert np.allclose(y_ext, scipy_y, rtol=1e-8)


def test_required_args():
    names = gauss_on_exgauss.required_args()
    assert names[0] == "mu"
    assert names[1] == "sigma"
    assert names[2] == "htail"
    assert names[3] == "tau"


def test_name():
    assert gauss_on_exgauss.name == "gauss_on_exgauss"



import numpy as np
from pytest import approx

import pygama.math.unbinned_fitting as pgubf


def test_fit_unbinned():
    xr = (-2, 2)  # xrange

    rng = np.random.default_rng(42)
    mu = 1
    sigma = 0.1

    xdata = rng.normal(mu, sigma, size=1000)
    xdata = xdata[(xr[0] < xdata) & (xdata < xr[1])]

    n, xe = np.histogram(xdata, bins=50, range=xr)
    from pygama.math.functions.gauss import gaussian

    fit, fit_error, fit_cov = pgubf.fit_unbinned(
        gaussian.get_pdf, xdata, guess=[0, 0.9], cost_func="LL", extended=False
    )
    assert fit["mu"] == approx(mu, rel=1e-2)
    assert fit["sigma"] == approx(sigma, rel=1e-1)



import numpy as np
from pytest import approx

import pygama.math.histogram as pgh


def test_get_hist():
    data = np.arange(-10, 10)
    hist, bins, err = pgh.get_hist(data, bins=20, range=(-10, 10))
    numpy_hist, numpy_bins = np.histogram(data, bins=20, range=(-10, 10))
    # Check that we agree with numpy
    assert hist == approx(numpy_hist, rel=1e-5)
    assert bins == approx(numpy_bins, rel=1e-5)
    assert err == approx(
        numpy_hist, rel=1e-5
    )  # for no weights, the variances should be the same as the histogram!
    # check that the values match what we expect
    assert np.array_equal(np.array(hist, dtype=np.float32), np.ones(len(data)))
    assert np.array_equal(np.array(bins, dtype=np.float32), np.arange(-10, 11))
    assert np.array_equal(np.array(err, dtype=np.float32), np.ones(len(data)))

    weight = np.full(len(data), 0.2)

    hist, bins, var = pgh.get_hist(data, bins=50, range=(-10, 10), dx=2, wts=weight)
    numpy_hist, numpy_bins = np.histogram(
        data, bins=10, range=(-10, 10), weights=weight
    )
    numpy_var, numpy_bins = np.histogram(
        data, bins=10, range=(-10, 10), weights=weight * weight
    )

    assert hist == approx(numpy_hist, rel=1e-5)
    assert bins == approx(numpy_bins, rel=1e-5)
    assert var == approx(numpy_var, rel=1e-5)

    assert np.allclose(
        hist, np.full(10, 2 * 0.2), rtol=1e-8
    )  # each bin has two elements, and each element has the 0.2 weight
    assert np.allclose(
        bins, np.array(np.arange(-10, 11, 2), dtype=np.float64), rtol=1e-8
    )
    assert np.allclose(var, np.full(10, 2 * 0.2 * 0.2), rtol=1e-8)


def test_better_int_binning():
    lo, hi, bins = pgh.better_int_binning(x_lo=-20.1, x_hi=10.1, n_bins=10.2)
    assert lo == -21
    assert hi == 12
    assert bins == 11


def test_get_bin_centers():
    bins = pgh.get_bin_centers(np.arange(-10.5, 10))
    assert np.array_equal(bins, np.arange(-10, 10))


def test_get_bin_widths():
    dx = pgh.get_bin_widths(np.arange(-10, 10))
    assert np.array_equal(dx, np.ones(len(dx)))


def test_find_bin():
    uniform_bins = np.arange(-10, 10)
    uniform_idx = pgh.find_bin(1.5, uniform_bins)
    assert uniform_idx == 11

    non_uniform_bins = np.array(
        [-1, -0.6, -0.3, 0, 0.1, 0.5, 0.76, 1, 1.1, 1.4, 1.6, 2], dtype=np.float32
    )
    non_uniform_idx = pgh.find_bin(1.5, non_uniform_bins)
    assert non_uniform_idx == 10


def test_range_slice():
    np.random.seed(42)
    data = np.random.normal(size=100).astype(np.float32)
    hist, bins, err = pgh.get_hist(data, bins=10, range=(-1, 1))

    hist_slice, bins_slice, var_slice = pgh.range_slice(0.1, 0.6, hist, bins, err)
    assert np.array_equal(hist_slice, [8.0, 13.0, 3.0])
    assert np.array_equal(
        np.array(bins_slice, dtype=np.float32),
        np.array([0.0, 0.2, 0.4, 0.6], dtype=np.float32),
    )
    assert np.array_equal(var_slice, [8.0, 13.0, 3.0])


def test_get_fwfm():
    from numpy.random import normal

    np.random.seed(42)
    hist, bins, var = pgh.get_hist(normal(size=10000), bins=100, range=(-5, 5))

    fwfm, dfwfm = pgh.get_fwfm(0.5, hist, bins, var, method="bins_over_f")
    assert fwfm == approx(2.4)
    assert dfwfm == approx(0.1911676414793058)

    fwfm, dfwfm = pgh.get_fwfm(0.5, hist, bins, var, method="interpolate")
    assert fwfm == approx(1.9897727272727268)
    assert dfwfm == approx(1.0740819016912582)

    fwfm, dfwfm = pgh.get_fwfm(0.5, hist, bins, var, method="fit_slopes")
    assert fwfm == approx(2.2320859865745684)
    assert dfwfm == approx(0.14298871443488284)


def test_get_fwhm():
    from numpy.random import normal

    np.random.seed(42)
    hist, bins, var = pgh.get_hist(normal(size=10000), bins=100, range=(-5, 5))

    fwfm, dfwfm = pgh.get_fwhm(hist, bins, var, method="bins_over_f")
    assert fwfm == approx(2.4)
    assert dfwfm == approx(0.1911676414793058)

    fwfm, dfwfm = pgh.get_fwhm(hist, bins, var, method="interpolate")
    assert fwfm == approx(1.9897727272727268)
    assert dfwfm == approx(1.0740819016912582)

    fwfm, dfwfm = pgh.get_fwhm(hist, bins, var, method="fit_slopes")
    assert fwfm == approx(2.2320859865745684)
    assert dfwfm == approx(0.14298871443488284)


def test_get_gaussian_guess():
    from numpy.random import normal

    np.random.seed(42)
    hist, bins, var = pgh.get_hist(normal(size=10000), bins=100, range=(-5, 5))

    gauss_params = pgh.get_gaussian_guess(hist, bins)
    assert np.allclose(
        np.array(gauss_params, dtype=np.float64),
        np.array(
            [-0.09999999999999964, 1.0191082802547773, 1047.3555083808512],
            dtype=np.float64,
        ),
        rtol=1e-8,
    )


def test_get_bin_estimates():
    from pygama.math.functions.gauss import nb_gauss

    par = np.array([1, 2])
    bins = np.arange(1, 10)
    estimates = pgh.get_bin_estimates(par, nb_gauss, bins, is_integral=False)

    assert np.allclose(
        np.array(estimates, dtype=np.float32),
        np.array(
            [
                9.69233234e-01,
                7.54839602e-01,
                4.57833362e-01,
                2.16265167e-01,
                7.95595087e-02,
                2.27941809e-02,
                5.08606923e-03,
                8.83826307e-04,
            ],
            dtype=np.float32,
        ),
        rtol=1e-8,
    )



"""
pygama convenience functions for 1D histograms.

1D hists in pygama require 3 things available from all implementations
of 1D histograms of numerical data in python: hist, bins, and var:
- hist: an array of histogram values
- bins: an array of bin edges
- var: an array of variances in each bin
If var is not provided, pygama assuems that the hist contains "counts" with
variance = counts (Poisson stats)

These are just convenience functions, provided for your convenience. Hopefully
they will help you if you need to do something trickier than is provided (e.g.
2D hists).
"""

import logging
from typing import Callable, Optional, Union

import hist as bh
import matplotlib.pyplot as plt
import numba as nb
import numpy as np
from matplotlib import rcParams

import pygama.math.utils as pgu

log = logging.getLogger(__name__)


def get_hist(
    data: np.ndarray,
    bins: Optional[Union[int, np.ndarray, str]] = None,
    range: Optional[tuple[float, float]] = None,
    dx: Optional[float] = None,
    wts: Optional[Union[float, np.ndarray]] = None,
) -> tuple[np.ndarray, ...]:
    """return hist, bins, var after binning data

    This is just a wrapper for humba.histogram, with optional weights for each
    element and proper computing of variances.

    Note: there are no overflow / underflow bins.

    Available binning methods:

      - Default (no binning arguments) : 100 bins over an auto-detected range
      - bins=n, range=(x_lo, x_hi) : n bins over the specified range (or leave
        range=None for auto-detected range)
      - bins=[str] : use one of np.histogram's automatic binning algorithms
      - bins=bin_edges_array : array lower bin edges, supports non-uniform binning
      - dx=dx, range=(x_lo, x_hi): bins of width dx over the specified range.
        Note: dx overrides the bins argument!

    Parameters
    ----------
    data
        The array of data to be histogrammed
    bins
        int: the number of bins to be used in the histogram
        array: an array of bin edges to use
        str: the name of the np.histogram automatic binning algorithm to use
        If not provided, humba.histogram's default auto-binning routine is used
    range
        (x_lo, x_high) is the tuple of low and high x values to uses for the
        very ends of the bin range. If not provided, np.histogram chooses the
        ends based on the data in data
    dx
        Specifies the bin width. Overrides "bins" if both arguments are present
    wts
        Array of weights for each bin. For example, if you want to divide all
        bins by a time T to get the bin contents in count rate, set wts = 1/T.
        Variances will be computed for each bin that appropriately account for
        each data point's weighting.

    Returns
    -------
    hist
        the values in each bin of the histogram
    bins
        an array of bin edges: bins[i] is the lower edge of the ith bin.
        Note: it includes the upper edge of the last bin and does not
        include underflow or overflow bins. So len(bins) = len(hist) + 1
    var
        array of variances in each bin of the histogram
    """
    if bins is None:
        bins = 100  # override boost_histogram.Histogram default of just 10

    if dx is not None:
        bins = int((range[1] - range[0]) / dx)

    if range is None:
        range = [np.amin(data), np.amax(data)]

    # bins includes left edge of first bin and right edge of all other bins
    # allow scalar weights
    if wts is not None and np.shape(wts) == ():
        wts = np.full_like(data, wts)

    # initialize the boost_histogram object
    if isinstance(bins, int):
        boost_histogram = bh.Hist(
            bh.axis.Regular(bins=bins, start=range[0], stop=range[1]),
            storage=bh.storage.Weight(),
        )
    else:
        # if bins are specified need to use variable
        boost_histogram = bh.Hist(bh.axis.Variable(bins), storage=bh.storage.Weight())
    # create the histogram
    boost_histogram.fill(data, weight=wts)
    # read out the histogram, bins, and variances
    hist, bins = boost_histogram.to_numpy()
    var = boost_histogram.variances()

    return hist, bins, var


def better_int_binning(
    x_lo: float = 0, x_hi: float = None, dx: float = None, n_bins: float = None
) -> tuple[int, int, int, int]:
    """Get a good binning for integer data.

    Guarantees an integer bin width.

    At least two of x_hi, dx, or n_bins must be provided.

    Parameters
    ----------
    x_lo
        Desired low x value for the binning
    x_hi
        Desired high x value for the binning
    dx
        Desired bin width
    n_bins
        Desired number of bins

    Returns
    -------
    x_lo
        int values for best x_lo
    x_hi
        int values for best x_hi, returned if x_hi is not None
    dx
        best int bin width, returned if arg dx is not None
    n_bins
        best int n_bins, returned if arg n_bins is not None
    """
    # process inputs
    n_nones = int(x_hi is None) + int(dx is None) + int(n_bins is None)
    if n_nones > 1:
        raise ValueError("must provide two of x_hi, dx or n_bins")
    if n_nones == 0:
        log.warning("Overconstrained. Ignoring x_hi.")
        x_hi = None

    # get valid dx or n_bins
    if dx is not None:
        if dx <= 0:
            raise ValueError(f"invalid dx={dx}")
        dx = np.round(dx)
        if dx == 0:
            dx = 1
    if n_bins is not None:
        if n_bins <= 0:
            raise ValueError(f"invalid n_bins={n_bins}")
        n_bins = np.round(n_bins)

    # can already return if no x_hi
    if x_hi is None:  # must have both dx and n_bins
        return int(x_lo), int(dx), int(n_bins)

    # x_hi is valid. Get a valid dx if we don't have one
    if dx is None:  # must have n_bins
        dx = np.round((x_hi - x_lo) / n_bins)
    if dx == 0:
        dx = 1

    # Finally, build a good binning from dx
    final_n_bins = np.ceil((x_hi - x_lo) / dx)
    x_lo = np.floor(x_lo)
    x_hi = x_lo + final_n_bins * dx
    if n_bins is None:
        return int(x_lo), int(x_hi), int(dx)
    else:
        return int(x_lo), int(x_hi), int(final_n_bins)


@nb.njit(parallel=False, fastmath=True)
def get_bin_centers(bins: np.ndarray) -> np.ndarray:
    """
    Returns an array of bin centers from an input array of bin edges.
    Works for non-uniform binning. Note: a new array is allocated

    Parameters
    ----------
    bins
        The input array of bin-edges

    Returns
    ----------
    bin_centers
        The array of bin centers
    """
    return (bins[:-1] + bins[1:]) / 2.0


@nb.njit(parallel=False, fastmath=True)
def get_bin_widths(bins: np.ndarray) -> np.ndarray:
    """
    Returns an array of bin widths from an input array of bin edges.
    Works for non-uniform binning.

    Parameters
    ----------
    bins
        The input array of bin-edges

    Returns
    ----------
    bin_widths
        An array of bin widths
    """
    return bins[1:] - bins[:-1]


@nb.njit(parallel=False, fastmath=True)
def find_bin(x: float, bins: np.ndarray) -> int:
    """
    Returns the index of the bin containing x
    Returns -1 for underflow, and len(bins) for overflow
    For uniform bins, jumps directly to the appropriate index.
    For non-uniform bins, binary search is used.

    Parameters
    ----------
    x
        The value to search for amongst the bins
    bins
        The input array of bin-edges

    Returns
    -------
    bin_idx
        Index of the bin containing x

    TODO: replace np.searchsorted with for loops, numba will speed this function up then
    """
    # first handle overflow / underflow
    if len(bins) == 0:
        return 0  # i.e. overflow
    if x < bins[0]:
        return -1
    if x > bins[-1]:
        return len(bins)

    # we are definitely in range, and there are at least 2 bin edges, one below
    # and one above x. try assuming uniform bins
    dx = bins[1] - bins[0]
    index = int(np.floor((x - bins[0]) / dx))
    if bins[index] <= x and bins[index + 1] > x:
        return index

    # bins are non-uniform: find by binary search
    return np.searchsorted(bins, x, side="right")


def range_slice(
    x_min: float,
    x_max: float,
    hist: np.ndarray,
    bins: np.ndarray,
    var: Optional[np.ndarray] = None,
) -> tuple[np.ndarray, ...]:
    """
    Get the histogram bins and values for a given slice of the range

    Parameters
    ----------
    x_min
        Lower endpoint of the range
    x_min
        Upper endpoint of the range
    hist, bins, var
        The histogrammed data to search through

    See Also
    --------
    find_bin
        for parameters and return values
    """
    i_min = find_bin(x_min, bins)
    i_max = find_bin(x_max, bins)
    if var is not None:
        var = var[i_min:i_max]
    return hist[i_min:i_max], bins[i_min : i_max + 1], var


def get_fwhm(
    hist: np.ndarray,
    bins: np.ndarray,
    var: Optional[np.ndarray] = None,
    mx: Optional[Union[float, tuple[float, float]]] = None,
    dmx: Optional[float] = 0,
    bl: Optional[Union[float, tuple[float, float]]] = 0,
    dbl: Optional[float] = 0,
    method: str = "bins_over_f",
    n_slope: int = 3,
) -> tuple[float, float]:
    """Convenience function for the FWHM of data in a histogram

    Typically used by sending slices around a peak. Searches about argmax(hist)
    for the peak to fall by [fraction] from mx to bl

    Parameters
    ----------
    fraction
        The fractional amplitude at which to evaluate the full width
    hist
        The histogram data array containing the peak
    bins
        An array of bin edges for the histogram
    var
        An array of histogram variances. Used with the 'fit_slopes' method
    mx
        The value to use for the max of the peak. If None, np.amax(hist) is
        used.
    dmx
        The uncertainty in mx
    bl
        Used to specify an offset from which to estimate the FWFM.
    dbl
        The uncertainty in the bl
    method
        'bins_over_f' : the simplest method: just take the difference in the bin
            centers that are over [fraction] of max. Only works for high stats and
            FWFM/bin_width >> 1
        'interpolate' : interpolate between the bins that cross the [fraction]
            line.  Works well for high stats and a reasonable number of bins.
            Uncertainty incorporates var, if provided.
        'fit_slopes' : fit over n_slope bins in the vicinity of the FWFM and
            interpolate to get the fractional crossing point. Works okay even
            when stats are moderate but requires enough bins that dx traversed
            by n_slope bins is approximately linear. Incorporates bin variances
            in fit and final uncertainties if provided.
    n_slope
        Number of bins in the vicinity of the FWFM used to interpolate the fractional
        crossing point with the 'fit_slopes' method

    Returns
    -------
    fwhm, dfwhm
        fwfm: the full width at half of the maximum above bl
        dfwfm: the uncertainty in fwhm


    See Also
    --------
    get_fwfm
        Function that computes the FWFM
    """
    if len(bins) == len(hist):
        log.warning(
            "note: this function has been updated to require bins rather than bin_centers. Don't trust this result"
        )
    return get_fwfm(0.5, hist, bins, var, mx, dmx, bl, dbl, method, n_slope)


def get_fwfm(
    fraction: float,
    hist: np.ndarray,
    bins: np.ndarray,
    var: Optional[np.ndarray] = None,
    mx: Optional[Union[float, tuple[float, float]]] = None,
    dmx: Optional[float] = 0,
    bl: Optional[Union[float, tuple[float, float]]] = 0,
    dbl: Optional[float] = 0,
    method: str = "bins_over_f",
    n_slope: int = 3,
) -> tuple[float, float]:
    """
    Estimate the full width at some fraction of the max of data in a histogram

    Typically used by sending slices around a peak. Searches about argmax(hist)
    for the peak to fall by [fraction] from mx to bl

    Parameters
    ----------
    fraction
        The fractional amplitude at which to evaluate the full width
    hist
        The histogram data array containing the peak
    bins
        An array of bin edges for the histogram
    var
        An array of histogram variances. Used with the 'fit_slopes' method
    mx
        The value to use for the max of the peak. If None, np.amax(hist) is
        used.
    dmx
        The uncertainty in mx
    bl
        Used to specify an offset from which to estimate the FWFM.
    dbl
        The uncertainty in the bl
    method
        'bins_over_f' : the simplest method: just take the difference in the bin
            centers that are over [fraction] of max. Only works for high stats and
            FWFM/bin_width >> 1
        'interpolate' : interpolate between the bins that cross the [fraction]
            line.  Works well for high stats and a reasonable number of bins.
            Uncertainty incorporates var, if provided.
        'fit_slopes' : fit over n_slope bins in the vicinity of the FWFM and
            interpolate to get the fractional crossing point. Works okay even
            when stats are moderate but requires enough bins that dx traversed
            by n_slope bins is approximately linear. Incorporates bin variances
            in fit and final uncertainties if provided.
    n_slope
        Number of bins in the vicinity of the FWFM used to interpolate the fractional
        crossing point with the 'fit_slopes' method

    Returns
    -------
    fwfm, dfwfm
        fwfm: the full width at [fraction] of the maximum above bl
        dfwfm: the uncertainty in fwfm

    Examples
    --------
    >>> import pygama.analysis.histograms as pgh
    >>> from numpy.random import normal
    >>> hist, bins, var = pgh.get_hist(normal(size=10000), bins=100, range=(-5,5))
    >>> pgh.get_fwfm(0.5, hist, bins, var, method='bins_over_f')
    (2.2, 0.15919638684132664) # may vary

    >>> pgh.get_fwfm(0.5, hist, bins, var, method='interpolate')
    (2.2041666666666666, 0.09790931254396479) # may vary

    >>> pgh.get_fwfm(0.5, hist, bins, var, method='fit_slopes')
    (2.3083363869003466, 0.10939486522749278) # may vary
    """
    # find bins over [fraction]
    if mx is None:
        mx = np.amax(hist)
        if var is not None and dmx == 0:
            dmx = np.sqrt(var[np.argmax(hist)])
    idxs_over_f = hist > (bl + fraction * (mx - bl))

    # argmax will return the index of the first occurrence of a maximum
    # so we can use it to find the first and last time idxs_over_f is "True"
    bin_lo = np.argmax(idxs_over_f)
    bin_hi = len(idxs_over_f) - np.argmax(idxs_over_f[::-1])
    bin_centers = get_bin_centers(bins)

    # precalc dheight: uncertainty in height used as the threshold
    dheight2 = (fraction * dmx) ** 2 + ((1 - fraction) * dbl) ** 2

    if method == "bins_over_f":
        # the simplest method: just take the difference in the bin centers
        fwfm = bin_centers[bin_hi] - bin_centers[bin_lo]

        # compute rough uncertainty as [bin width] (+) [dheight / slope]
        dx = bin_centers[bin_lo] - bin_centers[bin_lo - 1]
        dy = hist[bin_lo] - hist[bin_lo - 1]
        if dy == 0:
            dy = (hist[bin_lo + 1] - hist[bin_lo - 2]) / 3
        dfwfm2 = dx**2 + dheight2 * (dx / dy) ** 2
        dx = bin_centers[bin_hi + 1] - bin_centers[bin_hi]
        dy = hist[bin_hi] - hist[bin_hi + 1]
        if dy == 0:
            dy = (hist[bin_hi - 1] - hist[bin_hi + 2]) / 3
        dfwfm2 += dx**2 + dheight2 * (dx / dy) ** 2
        return fwfm, np.sqrt(dfwfm2)

    elif method == "interpolate":
        # interpolate between the two bins that cross the [fraction] line
        # works well for high stats
        if bin_lo < 1 or bin_hi >= len(hist) - 1:
            raise ValueError(f"Can't interpolate ({bin_lo}, {bin_hi})")

        val_f = bl + fraction * (mx - bl)

        # x_lo
        dx = bin_centers[bin_lo] - bin_centers[bin_lo - 1]
        dhf = val_f - hist[bin_lo - 1]
        dh = hist[bin_lo] - hist[bin_lo - 1]
        x_lo = bin_centers[bin_lo - 1] + dx * dhf / dh
        # uncertainty
        dx2_lo = 0
        if var is not None:
            dx2_lo = (dhf / dh) ** 2 * var[bin_lo] + ((dh - dhf) / dh) ** 2 * var[
                bin_lo - 1
            ]
            dx2_lo *= (dx / dh) ** 2
        dd_dh = -dx / dh

        # x_hi
        dx = bin_centers[bin_hi + 1] - bin_centers[bin_hi]
        dhf = hist[bin_hi] - val_f
        dh = hist[bin_hi] - hist[bin_hi + 1]
        if dh == 0:
            raise ValueError("Interpolation failed, dh == 0")
        x_hi = bin_centers[bin_hi] + dx * dhf / dh
        if x_hi < x_lo:
            raise ValueError("Interpolation produced negative fwfm")
        # uncertainty
        dx2_hi = 0
        if var is not None:
            dx2_hi = (dhf / dh) ** 2 * var[bin_hi + 1] + ((dh - dhf) / dh) ** 2 * var[
                bin_hi
            ]
            dx2_hi *= (dx / dh) ** 2
        dd_dh += dx / dh

        return x_hi - x_lo, np.sqrt(dx2_lo + dx2_hi + dd_dh**2 * dheight2)

    elif method == "fit_slopes":
        # evaluate the [fraction] point on a line fit to n_slope bins near the crossing.
        # works okay even when stats are moderate
        val_f = bl + fraction * (mx - bl)

        # x_lo
        i_0 = bin_lo - int(np.floor(n_slope / 2))
        if i_0 < 0:
            raise RuntimeError("Fit slopes failed")
        i_n = i_0 + n_slope
        wts = (
            None if var is None else 1 / np.sqrt(var[i_0:i_n])
        )  # fails for any var = 0
        wts = [w if w != np.inf else 0 for w in wts]

        try:
            (m, b), cov = np.polyfit(
                bin_centers[i_0:i_n], hist[i_0:i_n], 1, w=wts, cov="unscaled"
            )
        except np.linalg.LinAlgError:
            raise RuntimeError("LinAlgError in x_lo")
        x_lo = (val_f - b) / m
        # uncertainty
        dxl2 = (
            cov[0, 0] / m**2
            + (cov[1, 1] + dheight2) / (val_f - b) ** 2
            + 2 * cov[0, 1] / (val_f - b) / m
        )
        dxl2 *= x_lo**2

        # x_hi
        i_0 = bin_hi - int(np.floor(n_slope / 2)) + 1
        if i_0 == len(hist):
            raise RuntimeError("Fit slopes failed")

        i_n = i_0 + n_slope
        wts = None if var is None else 1 / np.sqrt(var[i_0:i_n])
        wts = [w if w != np.inf else 0 for w in wts]
        try:
            (m, b), cov = np.polyfit(
                bin_centers[i_0:i_n], hist[i_0:i_n], 1, w=wts, cov="unscaled"
            )
        except np.linalg.LinAlgError:
            raise RuntimeError("LinAlgError in x_hi")
        x_hi = (val_f - b) / m
        if x_hi < x_lo:
            raise RuntimeError("Fit slopes produced negative fwfm")

        # uncertainty
        dxh2 = (
            cov[0, 0] / m**2
            + (cov[1, 1] + dheight2) / (val_f - b) ** 2
            + 2 * cov[0, 1] / (val_f - b) / m
        )
        dxh2 *= x_hi**2

        return x_hi - x_lo, np.sqrt(dxl2 + dxh2)

    else:
        raise NameError(f"Unrecognized method {method}")


def plot_hist(
    hist: np.ndarray,
    bins: np.ndarray,
    var: Optional[np.ndarray] = None,
    show_stats: bool = False,
    stats_hloc: float = 0.75,
    stats_vloc: float = 0.85,
    fill: bool = False,
    fillcolor: str = "r",
    fillalpha: float = 0.2,
    **kwargs,
) -> None:
    """
    Plot a step histogram, with optional error bars

    Parameters
    ----------
    hist
        The histogram data array containing the peak
    bins
        An array of bin edges for the histogram
    var
        An array of histogram variances. Used with the 'fit_slopes' method
    show_stats
        If True, shows the mean, mean error, and standard deviation of the histogram on the plot
    stats_hloc
        matplotlib.pyplot horizontal location to place the stats
    stats_vloc
        matplotlib.pyplot vertical location to place the stats
    fill
        If True, fills in a step histogram when plotting
    fill_color
        Color to fill the step histogram if fill is True
    fill_alpha
        Alpha amount if fill is True
    """
    if fill:
        # the concat calls get the steps to draw correctly at the range boundaries
        # where="post" tells plt to draw the step y[i] between x[i] and x[i+1]
        save_color = None
        if "color" in kwargs:
            save_color = kwargs.pop("color")
        elif "c" in kwargs:
            save_color = kwargs.pop("c")
        plt.fill_between(
            np.concatenate(([bins[0]], bins)),
            np.concatenate(([0], hist, [0])),
            step="post",
            color=fillcolor,
            alpha=fillalpha,
            **kwargs,
        )
        if save_color is not None:
            kwargs["color"] = save_color
    if var is None:
        # the concat calls get the steps to draw correctly at the range boundaries
        # where="post" tells plt to draw the step y[i] between x[i] and x[i+1]
        plt.step(
            np.concatenate(([bins[0]], bins)),
            np.concatenate(([0], hist, [0])),
            where="post",
            **kwargs,
        )
    else:
        plt.errorbar(
            get_bin_centers(bins),
            hist,
            xerr=get_bin_widths(bins) / 2,
            yerr=np.sqrt(var),
            fmt="none",
            **kwargs,
        )
    if show_stats is True:
        bin_centers = get_bin_centers(bins)
        n = np.sum(hist)
        if n <= 1:
            raise RuntimeError(f"can't compute sigma for n = {n}")
        mean = np.sum(hist * bin_centers) / n
        x2ave = np.sum(hist * bin_centers * bin_centers) / n
        stddev = np.sqrt(n / (n - 1) * (x2ave - mean * mean))
        dmean = stddev / np.sqrt(n)

        mean, dmean = pgu.get_formatted_stats(mean, dmean, 2)
        stats = f"$\\mu={mean} \\pm {dmean}$\n$\\sigma={stddev:#.3g}$"
        stats_fontsize = rcParams["legend.fontsize"]
        plt.text(
            stats_hloc,
            stats_vloc,
            stats,
            transform=plt.gca().transAxes,
            fontsize=stats_fontsize,
        )


def get_gaussian_guess(
    hist: np.ndarray, bins: np.ndarray
) -> tuple[float, float, float]:
    """
    given a hist, gives guesses for mu, sigma, and amplitude

    Parameters
    ----------
    hist
        Array of histogrammed data
    bins
        Array of histogram bins

    Returns
    -------
    guess_mu
        guess for the mu parameter of a Gaussian
    guess_sigma
        guess for the sigma parameter of a Gaussian
    guess_area
        guess for the A parameter of a Gaussian
    """
    if len(bins) == len(hist):
        log.warning(
            "note: this function has been updated to require bins rather than bin_centers. Don't trust this result"
        )

    max_idx = np.argmax(hist)
    guess_mu = (bins[max_idx] + bins[max_idx]) / 2  # bin center
    guess_amp = hist[max_idx]

    # find 50% amp bounds on both sides for a FWHM guess
    guess_sigma = get_fwhm(hist, bins)[0] / 2.355  # FWHM to sigma
    guess_area = guess_amp * guess_sigma * np.sqrt(2 * np.pi)

    return (guess_mu, guess_sigma, guess_area)


def get_bin_estimates(
    pars: np.ndarray,
    func: Callable,
    bins: np.ndarray,
    is_integral: bool = False,
    **kwargs,
) -> np.ndarray:
    """
    Bin expected means are estimated by f(bin_center)*bin_width. Supply an
    integrating function to compute the integral over the bin instead.
    TODO: make default integrating function a numerical method that is off by
    default.

    Parameters
    ----------
    pars
        The parameters of the function, func
    func
        The function at which to evaluate the bin centers
    bins
        Array of histogram bins
    is_integral
        if True, then func is an integral function

    Returns
    -------
    f(bin_center)*bin_width
        The expected mean of a bin

    See Also
    --------
    get_bin_widths
        Returns the width of the bins
    get_bin_centers
        Finds the bin centers of the supplied bins
    """
    if is_integral is False:
        return func(get_bin_centers(bins), *pars, **kwargs) * get_bin_widths(bins)
    else:
        # func can be an integral functions
        return func(bins[1:], *pars, **kwargs) - func(bins[:-1], *pars, **kwargs)


def get_i_local_extrema(data, delta):
    """Get lists of indices of the local maxima and minima of data

    The "local" extrema are those maxima / minima that have heights / depths of
    at least delta.
    Converted from MATLAB script at: http://billauer.co.il/peakdet.html

    Parameters
    ----------
    data : array-like
        the array of data within which extrema will be found
    delta : scalar
        the absolute level by which data must vary (in one direction) about an
        extremum in order for it to be tagged

    Returns
    -------
    imaxes, imins : 2-tuple ( array, array )
        A 2-tuple containing arrays of variable length that hold the indices of
        the identified local maxima (first tuple element) and minima (second
        tuple element)
    """

    # prepare output
    imaxes, imins = [], []

    # sanity checks
    data = np.asarray(data)
    if not np.isscalar(delta):
        log.error("get_i_local_extrema: Input argument delta must be a scalar")
        return np.array(imaxes), np.array(imins)
    if delta <= 0:
        log.error(f"get_i_local_extrema: delta ({delta}) must be positive")
        return np.array(imaxes), np.array(imins)

    # now loop over data
    imax, imin = 0, 0
    find_max = True
    for i in range(len(data)):
        if data[i] > data[imax]:
            imax = i
        if data[i] < data[imin]:
            imin = i

        if find_max:
            # if the sample is less than the current max by more than delta,
            # declare the previous one a maximum, then set this as the new "min"
            if data[i] < data[imax] - delta:
                imaxes.append(imax)
                imin = i
                find_max = False
        else:
            # if the sample is more than the current min by more than delta,
            # declare the previous one a minimum, then set this as the new "max"
            if data[i] > data[imin] + delta:
                imins.append(imin)
                imax = i
                find_max = True

    return np.array(imaxes), np.array(imins)


def get_i_local_maxima(data, delta):
    return get_i_local_extrema(data, delta)[0]


def get_i_local_minima(data, delta):
    return get_i_local_extrema(data, delta)[1]



import numpy as np
from scipy.stats import norm

from pygama.math.functions.sum_dists import SumDists
from pygama.math.functions.triple_gauss_on_double_step import (
    triple_gauss_on_double_step,
)


def test_triple_gauss_on_double_step_pdf():
    x = np.arange(-50, 50)
    mu_1 = 0
    sigma_1 = 1
    hstep_1 = 1

    mu_2 = mu_1
    sigma_2 = sigma_1
    hstep_2 = hstep_1

    mu_3 = 5
    sigma_3 = 10

    x_lo = np.amin(x)
    x_hi = np.amax(x)
    n_sig1 = 2
    n_sig2 = 10
    n_sig3 = 4
    n_bkg1 = 4
    n_bkg2 = 3

    pars = np.array(
        [
            x_lo,
            x_hi,
            n_sig1,
            mu_1,
            sigma_1,
            n_sig2,
            mu_2,
            sigma_2,
            n_sig3,
            mu_3,
            sigma_3,
            n_bkg1,
            hstep_1,
            n_bkg2,
            hstep_2,
        ],
        dtype=float,
    )

    n_sig_array = [n_sig1, n_sig2, n_sig3, n_bkg1, n_bkg2]

    # compute the unnormalized step function
    scipy_y_step = n_bkg1 * (
        1 + hstep_1 * (norm.cdf(x, mu_1, sigma_1) * 2 - 1)
    ) + n_bkg2 * (
        1 + hstep_2 * (norm.cdf(x, mu_2, sigma_2) * 2 - 1)
    )  # pdf = (1+erf(x/np.sqrt(2)))/18; erf(x) = 2 cdf_norm(x*sqrt(2))-1

    maximum = (np.amax(x) - mu_1) / sigma_1
    minimum = (np.amin(x) - mu_1) / sigma_1
    normalization = sigma_1 * (
        (1 - hstep_1) * (maximum - minimum)
        + 2
        * hstep_1
        * (
            maximum * norm.cdf(maximum)
            + np.exp(-1 * maximum**2 / 2) / np.sqrt(2 * np.pi)
            - minimum * norm.cdf(minimum)
            - np.exp(-1 * minimum**2 / 2) / np.sqrt(2 * np.pi)
        )
    )
    scipy_y_step = scipy_y_step / normalization

    y_direct = triple_gauss_on_double_step.get_pdf(x, *pars)

    scipy_y_gauss = (
        n_sig1 * norm.pdf(x, mu_1, sigma_1)
        + n_sig2 * norm.pdf(x, mu_2, sigma_2)
        + n_sig3 * norm.pdf(x, mu_3, sigma_3)
    )

    scipy_y = scipy_y_step + scipy_y_gauss

    assert isinstance(triple_gauss_on_double_step, SumDists)

    assert np.allclose(y_direct, scipy_y, rtol=1e-8)

    y_sig, y_ext = triple_gauss_on_double_step.pdf_ext(x, *pars)
    assert np.allclose(y_ext, scipy_y, rtol=1e-8)
    assert np.allclose(
        y_sig, np.sum(n_sig_array), rtol=1e-4
    )  # We aren't using trun_norm, so this shouldn't be quite equal to the sum of the areas, but quite close


def test_triple_gauss_on_double_step_cdf():
    x = np.arange(-10, 10)
    mu_1 = 0
    sigma_1 = 1
    hstep_1 = 1

    mu_2 = 0
    sigma_2 = 1
    hstep_2 = 1

    mu_3 = 5
    sigma_3 = 10

    x_lo = np.amin(x)
    x_hi = np.amax(x)
    n_sig1 = 2
    n_sig2 = 10
    n_sig3 = 4
    n_bkg1 = 4
    n_bkg2 = 3

    pars = np.array(
        [
            x_lo,
            x_hi,
            n_sig1,
            mu_1,
            sigma_1,
            n_sig2,
            mu_2,
            sigma_2,
            n_sig3,
            mu_3,
            sigma_3,
            n_bkg1,
            hstep_1,
            n_bkg2,
            hstep_2,
        ],
        dtype=float,
    )

    y_direct = triple_gauss_on_double_step.get_cdf(x, *pars)
    scipy_y_step = n_bkg1 * (
        (x * (2 * norm.cdf(x, mu_1, sigma_1) - 1) / np.sqrt(2))
        + np.exp(-1 * x**2 / 2) / np.sqrt(np.pi)
        + x / np.sqrt(2)
    ) * 2 / 18 / np.sqrt(2) + n_bkg2 * (
        (x * (2 * norm.cdf(x, mu_2, sigma_2) - 1) / np.sqrt(2))
        + np.exp(-1 * x**2 / 2) / np.sqrt(np.pi)
        + x / np.sqrt(2)
    ) * 2 / 18 / np.sqrt(
        2
    )
    scipy_y_gauss = (
        n_sig1 * norm.cdf(x, mu_1, sigma_1)
        + n_sig2 * norm.cdf(x, mu_2, sigma_2)
        + n_sig3 * norm.cdf(x, mu_3, sigma_3)
    )
    scipy_y = scipy_y_step + scipy_y_gauss

    assert isinstance(triple_gauss_on_double_step, SumDists)

    assert np.allclose(y_direct, scipy_y, rtol=1e-8)

    y_ext = triple_gauss_on_double_step.cdf_ext(x, *pars)
    assert np.allclose(y_ext, scipy_y, rtol=1e-8)


def test_required_args():
    names = triple_gauss_on_double_step.required_args()
    assert names[0] == "x_lo"
    assert names[1] == "x_hi"
    assert names[2] == "n_sig1"
    assert names[3] == "mu1"
    assert names[4] == "sigma1"
    assert names[5] == "n_sig2"
    assert names[6] == "mu2"
    assert names[7] == "sigma2"
    assert names[8] == "n_sig3"
    assert names[9] == "mu3"
    assert names[10] == "sigma3"
    assert names[11] == "n_bkg1"
    assert names[12] == "hstep1"
    assert names[13] == "n_bkg2"
    assert names[14] == "hstep2"


def test_name():
    assert triple_gauss_on_double_step.name == "triple_gauss_on_double_step"



"""
pygama convenience functions for fitting hpge peak shape data
"""

import logging
import math
from typing import Optional

import numpy as np
from scipy.optimize import brentq, minimize_scalar

from pygama.math.functions.exgauss import nb_exgauss_pdf
from pygama.math.functions.gauss import nb_gauss_pdf
from pygama.math.functions.gauss_on_exgauss import gauss_on_exgauss
from pygama.math.functions.step import nb_unnorm_step_pdf

log = logging.getLogger(__name__)


def hpge_peak_fwhm(
    sigma: float, htail: float, tau: float, cov: Optional[float] = None
) -> tuple[float, float]:
    """
    Return the FWHM of the hpge_peak function, ignoring background and step
    components. If calculating error also need the normalisation for the step
    function.

    Parameters
    ----------
    sigma
        The width of the hpge_peak
    htail
        The height of the tail in the hpge_peak
    tau
        The characteristic scale in the extended Gaussian in the hpge_peak
    cov
        The covariant matrix of the previous parameters

    Returns
        FWHM, FWHM_uncertainty
            The FWHM of the hpge_peak and its uncertainty
    """

    # optimize this to find max value
    def neg_hpge_peak_peak_bgfree(e, sigma, htail, tau):
        return -gauss_on_exgauss.get_pdf(
            np.array([e]), *np.array([0, sigma, htail, tau])
        )[0]

    if htail < 0 or htail > 1:
        raise ValueError("htail outside allowed limits of 0 and 1")

    res = minimize_scalar(
        neg_hpge_peak_peak_bgfree,
        args=(sigma, htail, tau),
        bounds=(-sigma - htail, sigma + htail),
    )
    e_max = res.x
    half_max = -neg_hpge_peak_peak_bgfree(e_max, sigma, htail, tau) / 2.0

    # root find this to find the half-max energies
    def hpge_peak_peak_bgfree_halfmax(e, sigma, htail, tau, half_max):
        return (
            gauss_on_exgauss.get_pdf(np.array([e]), *np.array([0, sigma, htail, tau]))[
                0
            ]
            - half_max
        )

    try:
        lower_hm = brentq(
            hpge_peak_peak_bgfree_halfmax,
            -(2.5 * sigma / 2 + htail * tau),
            e_max,
            args=(sigma, htail, tau, half_max),
        )
    except Exception:
        lower_hm = brentq(
            hpge_peak_peak_bgfree_halfmax,
            -(5 * sigma + htail * tau),
            e_max,
            args=(sigma, htail, tau, half_max),
        )
    try:
        upper_hm = brentq(
            hpge_peak_peak_bgfree_halfmax,
            e_max,
            2.5 * sigma / 2,
            args=(sigma, htail, tau, half_max),
        )
    except Exception:
        upper_hm = brentq(
            hpge_peak_peak_bgfree_halfmax,
            e_max,
            5 * sigma,
            args=(sigma, htail, tau, half_max),
        )

    if cov is None:
        return upper_hm - lower_hm

    # calculate uncertainty
    # nsig set to 1, mu to 0, hstep+nbkg set to 0
    pars = [1, 0, sigma, htail, tau, 0, 0]
    step_norm = 1
    gradmax = hpge_peak_parameter_gradient(e_max, pars, step_norm)
    gradmax *= 0.5
    grad1 = hpge_peak_parameter_gradient(lower_hm, pars, step_norm)
    grad1 -= gradmax
    grad1 /= hpge_peak_peakshape_derivative(lower_hm, pars, step_norm)
    grad2 = hpge_peak_parameter_gradient(upper_hm, pars, step_norm)
    grad2 -= gradmax
    grad2 /= hpge_peak_peakshape_derivative(upper_hm, pars, step_norm)
    grad2 -= grad1

    fwfm_unc = np.sqrt(np.dot(grad2, np.dot(cov, grad2)))

    return upper_hm - lower_hm, fwfm_unc


def hpge_peak_fwfm(sigma, htail, tau, frac_max=0.5, cov=None):
    """
    Return the FWHM of the radford_peak function, ignoring background and step
    components. If calculating error also need the normalisation for the step
    function.
    """

    # optimize this to find max value
    def neg_radford_peak_bgfree(e, sigma, htail, tau):
        return -gauss_on_exgauss.get_pdf(np.array([e]), 0, sigma, htail, tau)[0]

    if htail < 0 or htail > 1:
        raise ValueError("htail outside allowed limits of 0 and 1")

    res = minimize_scalar(
        neg_radford_peak_bgfree,
        args=(sigma, htail, tau),
        bounds=(-sigma - htail, sigma + htail),
    )
    e_max = res.x
    val_frac_max = -neg_radford_peak_bgfree(e_max, sigma, htail, tau) * frac_max

    # root find this to find the half-max energies
    def radford_peak_bgfree_fracmax(e, sigma, htail, tau, val_frac_max):
        return (
            gauss_on_exgauss.get_pdf(np.array([e]), 0, sigma, htail, tau)[0]
            - val_frac_max
        )

    try:
        lower_hm = brentq(
            radford_peak_bgfree_fracmax,
            -(2.5 * sigma / 2 + htail * tau),
            e_max,
            args=(sigma, htail, tau, val_frac_max),
        )
    except Exception:
        lower_hm = brentq(
            radford_peak_bgfree_fracmax,
            -(5 * sigma + htail * tau),
            e_max,
            args=(sigma, htail, tau, val_frac_max),
        )
    try:
        upper_hm = brentq(
            radford_peak_bgfree_fracmax,
            e_max,
            2.5 * sigma / 2,
            args=(sigma, htail, tau, val_frac_max),
        )
    except Exception:
        upper_hm = brentq(
            radford_peak_bgfree_fracmax,
            e_max,
            5 * sigma,
            args=(sigma, htail, tau, val_frac_max),
        )

    if cov is None:
        return upper_hm - lower_hm
    # calculate uncertainty
    # nsig set to 1, mu to 0, hstep+nbkg set to 0
    pars = [1, 0, sigma, htail, tau, 0, 0]

    rng = np.random.default_rng(1)
    par_b = rng.multivariate_normal(pars, cov, size=100)
    y_b = np.zeros(len(par_b))
    for i, p in enumerate(par_b):
        try:
            y_b[i] = hpge_peak_fwfm(p[2], p[3], p[4], frac_max=frac_max)
        except Exception:
            y_b[i] = np.nan
    yerr_boot = np.nanstd(y_b, axis=0)

    return upper_hm - lower_hm, yerr_boot


def hpge_peak_mode(mu, sigma, htail, tau, cov=None):

    if htail < 0 or htail > 1:
        if cov is not None:
            return np.nan, np.nan
        else:
            return np.nan

    try:
        mode = brentq(
            hpge_peak_peakshape_derivative,
            mu - 2 * sigma - htail * tau,
            mu + 2 * sigma + htail * tau,
            args=([1, mu, sigma, htail, tau, 0, 0], 1),
        )
    except ValueError:
        try:
            mode = brentq(
                hpge_peak_peakshape_derivative,
                mu - 4 * sigma - htail * tau,
                mu + 4 * sigma + htail * tau,
                args=([1, mu, sigma, htail, tau, 0, 0], 1),
            )
        except ValueError:
            mode = np.nan

    if cov is None:
        return mode
    else:
        # nsig set to 1, hstep+nbkg set to 0
        pars = np.array([1, mu, sigma, htail, tau, 0, 0])
        rng = np.random.default_rng(1)
        par_b = rng.multivariate_normal(pars, cov, size=10000)
        modes = np.array([hpge_peak_mode(p[1], p[2], p[3], p[4]) for p in par_b])
        mode_err_boot = np.nanstd(modes, axis=0)

        return mode, mode_err_boot


def hpge_peak_peakshape_derivative(
    e: np.ndarray, pars: np.ndarray, step_norm: float
) -> np.ndarray:
    """
    Computes the derivative of the hpge_peak peak shape

    Parameters
    ----------
    e
        The array of energies of the hpge_peak
    pars
        The parameters of the hpge_peak fit
    step_norm
        The normalization of the background step function in the hpge_peak

    Returns
    -------
    derivative
        the derivative of the hpge_peak
    """
    n_sig, mu, sigma, htail, tau, n_bkg, hstep = pars

    sigma = abs(sigma)
    gaus = nb_gauss_pdf(e, mu, sigma)
    y = (e - mu) / sigma
    ret = -(1 - htail) * (y / sigma) * gaus
    ret -= (
        htail / tau * (-nb_exgauss_pdf(np.array([e, e - 1]), mu, sigma, tau)[0] + gaus)
    )

    return n_sig * ret - n_bkg * hstep * gaus / step_norm  # need norm factor for bkg


def hpge_peak_parameter_gradient(
    e: float, pars: np.ndarray, step_norm: float
) -> np.ndarray:
    """
    Computes the gradient of the hpge_peak parameters

    Parameters
    ----------
    e
        The energy of the hpge_peak
    pars
        The parameters of the hpge_peak fit
    step_norm
        The normalization of the background step function in the hpge_peak

    Returns
    -------
    gradient
        gradient of the n_sig, mu, sigma, h_tail, tau, n_bkg, and hstep parameters of the
        HPGe peak

    """
    n_sig, mu, sigma, htail, tau, n_bkg, hstep = pars

    gaus = nb_gauss_pdf(np.array([e, e - 1]), mu, sigma)[0]

    tail_l = nb_exgauss_pdf(np.array([e, e - 1]), mu, sigma, tau)[0]
    if n_bkg == 0:
        step_f = 0
    else:
        step_f = (
            nb_unnorm_step_pdf(np.array([e, e - 1]), mu, sigma, hstep)[0] / step_norm
        )

    # some unitless numbers that show up a bunch
    y = (e - mu) / sigma
    sig_tau_l = sigma / tau

    g_n_sig = 0.5 * (htail * tail_l + (1 - htail) * gaus)
    g_n_bkg = step_f

    g_hs = n_bkg * math.erfc(y / np.sqrt(2)) / step_norm

    g_ht = (n_sig / 2) * (tail_l - gaus)

    # gradient of gaussian part
    g_mu = (1 - htail) * y / sigma * gaus
    g_sigma = (1 - htail) * (y * y + -1) / sigma * gaus

    # gradient of low tail, use approximation if necessary
    g_mu += htail / tau * (-tail_l + gaus)
    g_sigma += htail / tau * (sig_tau_l * tail_l - (sig_tau_l - y) * gaus)
    g_tau = (
        -htail
        / tau
        * (
            (1.0 + sig_tau_l * y + sig_tau_l * sig_tau_l) * tail_l
            - sig_tau_l * sig_tau_l * gaus
        )
        * n_sig
    )

    g_mu = n_sig * g_mu + (2 * n_bkg * hstep * gaus) / step_norm
    g_sigma = n_sig * g_sigma + (2 * n_bkg * hstep * gaus * y) / (
        step_norm * np.sqrt(sigma)
    )

    gradient = g_n_sig, g_mu, g_sigma, g_ht, g_tau, g_n_bkg, g_hs
    return np.array(gradient)



import numpy as np
from scipy.stats import norm

from pygama.math.functions.gauss_on_step import gauss_on_step
from pygama.math.functions.sum_dists import SumDists


def test_gauss_on_step_pdf():
    x = np.arange(-10, 10)
    mu = 0.4
    sigma = 1.1
    hstep = 0.75
    x_lo = np.amin(x)
    x_hi = np.amax(x)
    n_sig = 2
    n_bkg = 4

    pars = np.array([x_lo, x_hi, n_sig, mu, sigma, n_bkg, hstep], dtype=float)

    y_direct = gauss_on_step.get_pdf(x, *pars)

    # compute the unnormalized step function
    scipy_step = 1 + hstep * (
        norm.cdf(x, mu, sigma) * 2 - 1
    )  # pdf = (1+erf(x/np.sqrt(2))); erf(x) = 2 cdf_norm(x*sqrt(2))-1

    # compute the normalization for the step function
    maximum = (np.amax(x) - mu) / sigma
    minimum = (np.amin(x) - mu) / sigma
    normalization = sigma * (
        (1 - hstep) * (maximum - minimum)
        + 2
        * hstep
        * (
            maximum * norm.cdf(maximum)
            + np.exp(-1 * maximum**2 / 2) / np.sqrt(2 * np.pi)
            - minimum * norm.cdf(minimum)
            - np.exp(-1 * minimum**2 / 2) / np.sqrt(2 * np.pi)
        )
    )

    # compute the normalized step function
    scipy_y_step = scipy_step / normalization

    scipy_y_gauss = norm.pdf(x, mu, sigma)
    scipy_y = n_sig * scipy_y_gauss + n_bkg * scipy_y_step

    assert isinstance(gauss_on_step, SumDists)
    assert np.allclose(y_direct, scipy_y, rtol=1e-8)

    y_sig, y_ext = gauss_on_step.pdf_ext(x, *pars)
    assert np.allclose(y_ext, scipy_y, rtol=1e-8)
    assert np.allclose(y_sig, n_sig + n_bkg, rtol=1e-8)

    y_sig = gauss_on_step.pdf_norm(x, *pars)
    assert np.allclose(y_ext, scipy_y, rtol=1e-8)


def test_gauss_on_step_cdf():
    x = np.arange(-10, 10)
    mu = 0.56
    sigma = 1.009
    hstep = 0.753
    x_lo = np.amin(x)
    x_hi = np.amax(x)
    n_sig = 2
    n_bkg = 4

    pars = np.array([x_lo, x_hi, n_sig, mu, sigma, n_bkg, hstep], dtype=float)

    y_direct = gauss_on_step.get_cdf(x, *pars)

    # Compute the normalization of the pdf
    maximum = (np.amax(x) - mu) / sigma
    minimum = (np.amin(x) - mu) / sigma
    pdf_normalization = sigma * (
        (1 - hstep) * (maximum - minimum)
        + 2
        * hstep
        * (
            maximum * norm.cdf(maximum)
            + np.exp(-1 * maximum**2 / 2) / np.sqrt(2 * np.pi)
            - minimum * norm.cdf(minimum)
            - np.exp(-1 * minimum**2 / 2) / np.sqrt(2 * np.pi)
        )
    )

    z = (x - mu) / sigma
    # Compute the unnormalized cdf
    unnormalized_cdf = sigma * (
        (1 - hstep) * (z - minimum)
        + 2
        * hstep
        * (
            z * norm.cdf(z)
            + np.exp(-1 * z**2 / 2) / np.sqrt(2 * np.pi)
            - minimum * norm.cdf(minimum)
            - np.exp(-1 * minimum**2 / 2) / np.sqrt(2 * np.pi)
        )
    )

    # Compute the cdf
    scipy_y_step = unnormalized_cdf / pdf_normalization

    scipy_y_gauss = norm.cdf(x, mu, sigma)
    # for the cdf, we actually do not want to normalize by the sum of the areas, this is because iminuit takes a total unnormalized cdf for extended binned fits
    scipy_y = n_sig * scipy_y_gauss + n_bkg * scipy_y_step

    assert isinstance(gauss_on_step, SumDists)

    assert np.allclose(y_direct, scipy_y, rtol=1e-8)

    y_ext = gauss_on_step.cdf_ext(x, *pars)
    assert np.allclose(y_ext, scipy_y, rtol=1e-8)

    y_norm = gauss_on_step.cdf_norm(x, *pars)
    scipy_y_norm = (1 / (n_sig + n_bkg)) * (
        n_sig * scipy_y_gauss + n_bkg * scipy_y_step
    )

    assert np.allclose(y_norm, scipy_y_norm, rtol=1e-8)


def test_required_args():
    names = gauss_on_step.required_args()
    assert names[0] == "x_lo"
    assert names[1] == "x_hi"
    assert names[2] == "n_sig"
    assert names[3] == "mu"
    assert names[4] == "sigma"
    assert names[5] == "n_bkg"
    assert names[6] == "hstep"


def test_name():
    assert gauss_on_step.name == "gauss_on_step"



"""
This module contains the functions for performing the filter optimisation.
This happens with a grid search performed on ENC peak.
"""

import logging
import time

import lgdo
import matplotlib.pyplot as plt
import numpy as np
import scipy.stats
from scipy.interpolate import BSpline, splev, splrep
from scipy.optimize import minimize

from pygama.math.binned_fitting import goodness_of_fit
from pygama.math.distributions import gauss_on_uniform
from pygama.math.histogram import get_hist
from pygama.math.unbinned_fitting import fit_unbinned
from pygama.pargen.dsp_optimize import run_one_dsp

log = logging.getLogger(__name__)


def noise_optimization(
    tb_data: lgdo.Table,
    dsp_proc_chain: dict,
    par_dsp: dict,
    opt_dict: dict,
    lh5_path: str,
    display: int = 0,
) -> dict:
    """
    This function calculates the optimal filter par.
    Parameters
    ----------
    tb_data : str
        raw table to run the macro on
    dsp_proc_chain: str
        Path to minimal dsp config file
    par_dsp: str
        Dictionary with default dsp parameters
    opt_dict: str
        Dictionary with parameters for optimization
    lh5_path:  str
        Name of channel to process, should be name of lh5 group in raw files
    Returns
    -------
    res_dict : dict
    """

    t0 = time.time()

    samples = np.arange(opt_dict["start"], opt_dict["stop"], opt_dict["step"])
    samples_val = np.arange(opt_dict["start"], opt_dict["stop"], opt_dict["step_val"])

    opt_dict_par = opt_dict["optimization"]

    res_dict = {}
    if display > 0:
        dsp_data = run_one_dsp(tb_data, dsp_proc_chain, db_dict=par_dsp)
        psd = np.mean(dsp_data["wf_psd"].values.nda, axis=0)
        sample_us = float(dsp_data["wf_presum"].dt.nda[0]) / 1000
        freq = np.linspace(0, (1 / sample_us) / 2, len(psd))
        fig, ax = plt.subplots(figsize=(12, 6.75), facecolor="white")
        ax.plot(freq, psd)
        ax.set_xscale("log")
        ax.set_yscale("log")
        ax.set_xlabel("frequency (MHz)")
        ax.set_ylabel("power spectral density")

        plot_dict = {}
        plot_dict["nopt"] = {"fft": {"frequency": freq, "psd": psd, "fig": fig}}
        plt.close()

    result_dict = {}
    ene_pars = [par for par in opt_dict_par.keys()]
    log.info(f"\nRunning optimization for {ene_pars}")
    for i, x in enumerate(samples):
        x = f"{x:.1f}"
        log.info(f"\nCase {i}, par = {x} us")
        for ene_par in ene_pars:
            dict_str = opt_dict_par[ene_par]["dict_str"]
            filter_par = opt_dict_par[ene_par]["filter_par"]
            if dict_str in par_dsp:
                par_dsp[dict_str].update({filter_par: f"{x}*us"})
            else:
                par_dsp[dict_str] = {filter_par: f"{x}*us"}

        t1 = time.time()
        dsp_data = run_one_dsp(tb_data, dsp_proc_chain, db_dict=par_dsp)
        log.info(f"Time to process dsp data {time.time()-t1:.2f} s")

        for ene_par in ene_pars:
            dict_str = opt_dict_par[ene_par]["dict_str"]
            ene_str = opt_dict_par[ene_par]["ene_str"]
            if dict_str not in result_dict:
                result_dict[dict_str] = {}
            par_dict_res = result_dict[dict_str]

            energies = dsp_data[ene_str].nda

            if opt_dict["perform_fit"]:
                fom_results = simple_gaussian_fit(energies, dx=opt_dict["dx"])
            else:
                fom_results = calculate_spread(
                    energies,
                    opt_dict["percentile_low"],
                    opt_dict["percentile_high"],
                    opt_dict["n_bootstrap_samples"],
                )

            par_dict_res[x] = {}
            par_dict_res[x]["energies"] = energies
            par_dict_res[x]["fom"] = fom_results["fom"]
            par_dict_res[x]["fom_err"] = fom_results["fom_err"]

    for ene_par in ene_pars:
        log.info(f"\nOptimization for {ene_par}")
        dict_str = opt_dict_par[ene_par]["dict_str"]
        par_dict_res = result_dict[dict_str]
        sample_list = np.array([float(x) for x in result_dict[dict_str].keys()])
        fom_list = np.array(
            [result_dict[dict_str][x]["fom"] for x in result_dict[dict_str].keys()]
        )
        fom_err_list = np.array(
            [result_dict[dict_str][x]["fom_err"] for x in result_dict[dict_str].keys()]
        )

        guess_par = sample_list[np.nanargmin(fom_list)]

        tck = splrep(sample_list, fom_list, k=opt_dict["fit_deg"])
        tck = BSpline(tck[0], tck[1], tck[2])

        result = minimize(splev, guess_par, args=(tck))
        best_par = result.x[0]
        if (best_par < np.min(sample_list)) or (best_par > np.max(sample_list)):
            log.info(
                f"Par from minimization not accepted {best_par:.2f}, setting par to guess"
            )
            best_par = guess_par

        best_val = splev(best_par, tck)

        b_best_pars = np.zeros(opt_dict["n_bootstrap_samples"])
        for i in range(opt_dict["n_bootstrap_samples"]):
            indices = np.random.choice(len(sample_list), len(sample_list), replace=True)
            b_sample_list = sample_list[indices]
            b_fom_list = fom_list[indices]
            b_best_pars[i] = b_sample_list[np.nanargmin(b_fom_list)]
        best_par_err = np.std(b_best_pars)
        log.info(f"best par: {best_par:.2f} ± {best_par_err:.2f} us")

        par_dict_res["best_par"] = best_par
        par_dict_res["best_par_err"] = best_par_err
        par_dict_res["best_val"] = best_val

        filter_par = opt_dict_par[ene_par]["filter_par"]
        res_dict[dict_str] = {
            filter_par: f"{best_par:.2f}*us",
            f"{filter_par}_err": f"{best_par_err:.2f}*us",
        }

        if display > 0:
            plot_range = opt_dict["plot_range"]
            fig, ax = plt.subplots(figsize=(12, 6.75), facecolor="white")
            for i, x in enumerate(sample_list):
                x = f"{x:.1f}"
                energies = par_dict_res[x]["energies"]
                par_dict_res[x].pop("energies")
                hist, bins, var = get_hist(
                    energies, range=plot_range, dx=opt_dict["dx"]
                )
                bc = (bins[:-1] + bins[1:]) / 2.0
                string_res = (
                    f"par = {x} us, FOM = {fom_list[i]:.3f} ± {fom_err_list[i]:.3f} ADC"
                )
                ax.plot(bc, hist, ds="steps", label=string_res)
                log.info(string_res)
            ax.set_xlabel("energy (ADC)")
            ax.set_ylabel("counts")
            ax.legend(loc="upper right")
            par_dict_res["distribution"] = fig
            if display > 1:
                plt.show()
            else:
                plt.close()

            fig, ax = plt.subplots(figsize=(12, 6.75), facecolor="white")
            ax.errorbar(
                sample_list,
                fom_list,
                yerr=fom_err_list,
                color="b",
                fmt="x",
                ms=4,
                ls="",
                capsize=4,
                label="samples",
            )
            ax.plot(samples_val, splev(samples_val, tck), "k:", label="fit")
            ax.errorbar(
                best_par,
                best_val,
                xerr=best_par_err,
                color="r",
                fmt="o",
                ms=6,
                ls="",
                capsize=4,
                label=rf"best par: {best_par:.2f} ± {best_par_err:.2f} $\mu$s",
            )
            ax.set_xlabel(rf"{ene_par} parameter ($\mu$s)")
            ax.set_ylabel("FOM (ADC)")
            ax.legend()
            if display > 1:
                plt.show()
            else:
                plt.close()
            par_dict_res["optimization"] = fig
            plot_dict["nopt"][dict_str] = par_dict_res

    log.info(f"Time to complete the optimization {time.time()-t0:.2f} s")
    if display > 0:
        return res_dict, plot_dict
    else:
        return res_dict


def calculate_spread(energies, percentile_low, percentile_high, n_samples):
    spreads = np.zeros(n_samples)
    for i in range(n_samples):
        resampled = np.random.choice(energies, size=len(energies), replace=True)
        spread = np.percentile(resampled, percentile_high) - np.percentile(
            resampled, percentile_low
        )
        spreads[i] = spread

    mean_spread = np.mean(spreads)
    std_spread = np.std(spreads, ddof=1) / np.sqrt(n_samples)

    results = {}
    results["fom"] = mean_spread
    results["fom_err"] = std_spread
    return results


def simple_gaussian_fit(energies, dx=1, sigma_thr=4, allowed_p_val=1e-20):
    fit_range = [np.percentile(energies, 0.2), np.percentile(energies, 99.8)]

    hist, bins, var = get_hist(energies, range=fit_range, dx=dx)
    guess, bounds = simple_gaussian_guess(hist, bins, gauss_on_uniform)
    fit_range = [guess[0] - sigma_thr * guess[1], guess[0] + sigma_thr * guess[1]]

    energies_fit = energies[(energies > fit_range[0]) & (energies < fit_range[1])]
    pars, errs, cov = fit_unbinned(
        gauss_on_uniform.pdf_ext,
        energies_fit,
        guess=guess,
        bounds=bounds,
    )

    mu, mu_err = pars[0], errs[0]
    fwhm = pars[1] * 2 * np.sqrt(2 * np.log(2))
    fwhm_err = errs[1] * 2 * np.sqrt(2 * np.log(2))

    hist, bins, var = get_hist(energies_fit, range=fit_range, dx=dx)
    gof_pars = pars
    gof_pars[2] *= dx
    chisq, dof = goodness_of_fit(
        hist, bins, None, gauss_on_uniform.pdf_norm, gof_pars, method="Pearson"
    )
    p_val = scipy.stats.chi2.sf(chisq, dof + len(gof_pars))

    if (
        sum(sum(c) if c is not None else 0 for c in cov[2:, :][:, 2:]) == np.inf
        or sum(sum(c) if c is not None else 0 for c in cov[2:, :][:, 2:]) == 0
        or np.isnan(sum(sum(c) if c is not None else 0 for c in cov[2:, :][:, 2:]))
    ):
        log.debug("fit failed, cov estimation failed")
        fit_failed = True
    elif (np.abs(np.array(errs)[2:] / np.array(pars)[2:]) < 1e-7).any() or np.isnan(
        np.array(errs)[2:]
    ).any():
        log.debug("fit failed, parameter error too low")
        fit_failed = True
    elif p_val < allowed_p_val or np.isnan(p_val):
        log.debug("fit failed, parameter error too low")
        fit_failed = True
    else:
        fit_failed = False

    if fit_failed:
        log.debug("Returning values from guess")
        mu = guess[0]
        mu_err = 0
        fwhm = guess[1] * 2 * np.sqrt(2 * np.log(2))
        fwhm_err = 0

    results = {
        "pars": pars,
        "errors": errs,
        "covariance": cov,
        "mu": mu,
        "mu_err": mu_err,
        "fom": fwhm,
        "fom_err": fwhm_err,
        "chisq": chisq / dof,
        "p_val": p_val,
    }
    return results


def simple_gaussian_guess(hist, bins, func, toll=0.2):
    max_idx = np.argmax(hist)
    mu = bins[max_idx]
    max_amp = np.max(hist)

    idx = np.where(hist > max_amp / 2)
    ilo, ihi = idx[0][0], idx[0][-1]

    sigma = (bins[ihi] - bins[ilo]) / 2.355

    if sigma == 0:
        log.debug("error in sigma evaluation, using 2*(bin width) as sigma")
        sigma = 2 * (bins[1] - bins[0])

    dx = np.diff(bins)[0]
    n_bins_range = int((4 * sigma) // dx)

    min_idx = max_idx - n_bins_range
    max_idx = max_idx + n_bins_range
    min_idx = max(0, min_idx)
    max_idx = min(len(hist), max_idx)

    n_sig = np.sum(hist[min_idx:max_idx])

    guess = {"mu": mu, "sigma": sigma, "n_sig": n_sig}
    bounds = {
        "mu": (mu - sigma, mu + sigma),
        "sigma": (sigma - sigma * toll, sigma + sigma * toll),
        "n_sig": (n_sig + n_sig * toll, n_sig + n_sig * toll),
    }

    for par in func.required_args():
        if par == "x_lo" or par == "x_hi":
            guess[par] = np.inf
            bounds[par] = None
        elif par == "n_bkg" or par == "hstep":
            guess[par] = 0
            bounds[par] = None
    return guess, bounds



import numpy as np

import pygama.math.utils as pgu


def test_sizeof_fmt():
    size = pgu.sizeof_fmt(4012)
    assert size == "3.918 KB"


def test_get_par_names():
    names = pgu.get_par_names(pgu.sizeof_fmt)
    assert names == ["suffix"]
    names = pgu.get_par_names(pgu.print_fit_results)
    assert names == ["cov", "func", "title", "pad"]


def test_get_formatted_stats():
    stats = pgu.get_formatted_stats(10.009, 10.009, ndigs=4)
    assert stats == ("10.01", "10.01")


def test_print_fit_results(caplog):
    pgu.print_fit_results(
        np.ones((3,)), np.ones((3, 3)), func=None, title="Test", pad=True
    )
    assert [
        "Test:",
        "p0 = 1.0 +/- 1.0",
        "p1 = 1.0 +/- 1.0",
        "p2 = 1.0 +/- 1.0",
        "",
    ] == [rec.message for rec in caplog.records]



"""
Poisson distributions for pygama
"""

import numba as nb
import numpy as np
from scipy.stats import rv_discrete

from pygama.utils import numba_math_defaults as nb_defaults
from pygama.utils import numba_math_defaults_kwargs as nb_kwargs


@nb.njit(**nb_defaults(parallel=False))
def factorial(nn):
    res = 1
    for ii in nb.prange(2, nn + 1):
        res *= ii
    return res


@nb.njit(**nb_kwargs)
def nb_poisson_pmf(x: np.ndarray, mu: int, lamb: float) -> np.ndarray:
    r"""
    Normalised Poisson distribution, w/ args: mu, lamb.
    The range of support is :math:`\mathbb{N}`, with :math:`lamb` :math:`\in (0,\infty)`, :math:`\mu \in \mathbb{N}`
    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.

    .. math::
        pmf(x, \lambda, \mu) = \frac{\lambda^{x-\mu} e^{-\lambda}}{(x-\mu)!}

    Parameters
    ----------
    x : integer array-like
        The input data
    mu
        Amount to shift the distribution
    lamb
        The rate
    """

    y = np.empty_like(x, dtype=np.float64)
    for i in nb.prange(x.shape[0]):
        y[i] = x[i] - mu
        if y[i] < 0:
            y[i] = 0
        else:
            y[i] = lamb ** y[i] * np.exp(-lamb) / factorial(int(y[i]))
    return y


@nb.njit(**nb_kwargs)
def nb_poisson_cdf(x: np.ndarray, mu: int, lamb: float) -> np.ndarray:
    r"""
    Normalised Poisson cumulative distribution, w/ args: mu, lamb.
    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.

    .. math::
        cdf(x, \lambda, \mu) = e^{-\lambda}\sum_{j=0}^{\lfloor x-\mu \rfloor}\frac{\lambda^j}{j!}

    Parameters
    ----------
    x : integer array-like
        The input data
    mu
        Amount to shift the distribution
    lamb
        The rate
    """

    y = np.empty_like(x, dtype=np.float64)
    for i in nb.prange(x.shape[0]):
        y[i] = x[i] - mu
        z = 0
        for j in nb.prange(1, np.floor(y[i]) + 2):
            j -= 1
            z += lamb**j / factorial(j)
        y[i] = z * np.exp(-lamb)
    return y


@nb.njit(**nb_defaults(parallel=False))
def nb_poisson_scaled_pmf(
    x: np.ndarray, area: float, mu: int, lamb: float
) -> np.ndarray:
    r"""
    Scaled Poisson probability distribution, w/ args: lamb, mu.
    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.

    Parameters
    ----------
    x : integer array-like
        The input data
    area
        The number of counts in the signal
    mu
        Amount to shift the distribution
    lamb
        The rate
    """

    return area * nb_poisson_pmf(x, mu, lamb)


@nb.njit(**nb_defaults(parallel=False))
def nb_poisson_scaled_cdf(
    x: np.ndarray, area: float, mu: int, lamb: float
) -> np.ndarray:
    r"""
    Poisson cdf scaled by the number of signal counts for extended binned fits
    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.

    Parameters
    ----------
    x : integer array-like
        The input data
    area
        The number of counts in the signal
    mu
        Amount to shift the distribution
    lamb
        The rate
    """

    return area * nb_poisson_cdf(x, mu, lamb)


class PoissonGen(rv_discrete):

    def __init__(self, *args, **kwargs):
        self.x_lo = 0
        self.x_hi = np.inf
        super().__init__(*args, **kwargs)

    def _pmf(self, x: np.array, mu: int, lamb: float) -> np.array:
        x.flags.writeable = True
        return nb_poisson_pmf(x, mu[0], lamb[0])

    def _cdf(self, x: np.array, mu: int, lamb: float) -> np.array:
        x.flags.writeable = True
        return nb_poisson_cdf(x, mu[0], lamb[0])

    def get_pmf(self, x: np.array, mu: int, lamb: float) -> np.array:
        return nb_poisson_pmf(x, mu, lamb)

    def get_cdf(self, x: np.array, mu: int, lamb: float) -> np.array:
        return nb_poisson_cdf(x, mu, lamb)

    def pmf_ext(
        self, x: np.array, x_lo: float, x_hi: float, area: float, mu: int, lamb: float
    ) -> np.array:
        return np.diff(nb_poisson_scaled_cdf(np.array([x_lo, x_hi]), area, mu, lamb))[
            0
        ], nb_poisson_scaled_pmf(x, area, mu, lamb)

    def cdf_ext(self, x: np.array, area: float, mu: int, lamb: float) -> np.array:
        return nb_poisson_scaled_cdf(x, area, mu, lamb)

    def required_args(self) -> tuple[str, str]:
        return "mu", "lamb"


poisson = PoissonGen(a=0, name="poisson")
